import pkg_resources

import numpy as np
from scipy import interpolate

import astropy.units as u
from astropy.table import Table
from astropy.modeling.models import Drude1D, Polynomial1D, PowerLaw1D

from .baseclasses import BaseExtRvModel, BaseExtRvAfAModel
from .helpers import _get_x_in_wavenumbers, _test_valid_x_range, _smoothstep
from .averages import G03_SMCBar
from .shapes import _curve_F99_method, _modified_drude, FM90

# fmt: off
__all__ = ["CCM89", "O94", "F99", "F04", "VCG04", "GCC09", "M14", "G16", "F19",
           "D22", "G23"]
# fmt: on

x_range_CCM89 = [0.3, 10.0]
x_range_O94 = x_range_CCM89
x_range_F99 = [0.3, 10.0]
x_range_F04 = [0.3, 10.0]
x_range_VCG04 = [3.3, 8.0]
x_range_GCC09 = [3.3, 11.0]
x_range_M14 = [0.3, 3.3]
x_range_G16 = [0.3, 10.0]
x_range_G23 = [1.0 / 32.0, 1.0 / 0.0912]


class CCM89(BaseExtRvModel):
    r"""
    Cardelli, Clayton, & Mathis (1989) Milky Way R(V) dependent model

    Parameters
    ----------
    Rv: float
        R(V) = A(V)/E(B-V) = total-to-selective extinction

    Raises
    ------
    InputParameterError
       Input Rv values outside of defined range

    Notes
    -----
    From Cardelli, Clayton, and Mathis (1989, ApJ, 345, 245)

    Example showing CCM89 curves for a range of R(V) values.

    .. plot::
        :include-source:

        import numpy as np
        import matplotlib.pyplot as plt
        import astropy.units as u

        from dust_extinction.parameter_averages import CCM89

        fig, ax = plt.subplots()

        # generate the curves and plot them
        x = np.arange(0.5,10.0,0.1)/u.micron

        Rvs = ['2.0','3.0','4.0','5.0','6.0']
        for cur_Rv in Rvs:
           ext_model = CCM89(Rv=cur_Rv)
           ax.plot(x,ext_model(x),label='R(V) = ' + str(cur_Rv))

        ax.set_xlabel(r'$x$ [$\mu m^{-1}$]')
        ax.set_ylabel(r'$A(x)/A(V)$')

        ax.legend(loc='best')
        plt.show()
    """

    Rv_range = [2.0, 6.0]
    x_range = x_range_CCM89

    @staticmethod
    def evaluate(in_x, Rv):
        """
        CCM89 function

        Parameters
        ----------
        in_x: float
           expects either x in units of wavelengths or frequency
           or assumes wavelengths in wavenumbers [1/micron]

           internally wavenumbers are used

        Returns
        -------
        axav: np array (float)
            A(x)/A(V) extinction curve [mag]

        Raises
        ------
        ValueError
           Input x values outside of defined range
        """
        x = _get_x_in_wavenumbers(in_x)

        # check that the wavenumbers are within the defined range
        _test_valid_x_range(x, x_range_CCM89, "CCM89")

        # setup the a & b coefficient vectors
        n_x = len(x)
        a = np.zeros(n_x)
        b = np.zeros(n_x)

        # define the ranges
        ir_indxs = np.where(np.logical_and(0.3 <= x, x < 1.1))
        opt_indxs = np.where(np.logical_and(1.1 <= x, x < 3.3))
        nuv_indxs = np.where(np.logical_and(3.3 <= x, x <= 8.0))
        fnuv_indxs = np.where(np.logical_and(5.9 <= x, x <= 8))
        fuv_indxs = np.where(np.logical_and(8 < x, x <= 10))

        # Infrared
        y = x[ir_indxs] ** 1.61
        a[ir_indxs] = 0.574 * y
        b[ir_indxs] = -0.527 * y

        # NIR/optical
        y = x[opt_indxs] - 1.82
        a[opt_indxs] = np.polyval(
            (0.32999, -0.7753, 0.01979, 0.72085, -0.02427, -0.50447, 0.17699, 1), y
        )
        b[opt_indxs] = np.polyval(
            (-2.09002, 5.3026, -0.62251, -5.38434, 1.07233, 2.28305, 1.41338, 0), y
        )

        # NUV
        a[nuv_indxs] = (
            1.752 - 0.316 * x[nuv_indxs] - 0.104 / ((x[nuv_indxs] - 4.67) ** 2 + 0.341)
        )
        b[nuv_indxs] = (
            -3.09 + 1.825 * x[nuv_indxs] + 1.206 / ((x[nuv_indxs] - 4.62) ** 2 + 0.263)
        )

        # far-NUV
        y = x[fnuv_indxs] - 5.9
        a[fnuv_indxs] += -0.04473 * (y**2) - 0.009779 * (y**3)
        b[fnuv_indxs] += 0.2130 * (y**2) + 0.1207 * (y**3)

        # FUV
        y = x[fuv_indxs] - 8.0
        a[fuv_indxs] = np.polyval((-0.070, 0.137, -0.628, -1.073), y)
        b[fuv_indxs] = np.polyval((0.374, -0.42, 4.257, 13.67), y)

        # return A(x)/A(V)
        return a + b / Rv


class O94(BaseExtRvModel):
    r"""
    O'Donnell (1994) Milky Way R(V) dependent model

    Parameters
    ----------
    Rv: float
        R(V) = A(V)/E(B-V) = total-to-selective extinction

    Raises
    ------
    InputParameterError
       Input Rv values outside of defined range

    Notes
    -----
    From O'Donnell (1994, ApJ, 422, 158)
      Updates/improves the optical portion of the CCM89 model

    Example showing O94 curves for a range of R(V) values.

    .. plot::
        :include-source:

        import numpy as np
        import matplotlib.pyplot as plt
        import astropy.units as u

        from dust_extinction.parameter_averages import O94

        fig, ax = plt.subplots()

        # generate the curves and plot them
        x = np.arange(0.5,10.0,0.1)/u.micron

        Rvs = ['2.0','3.0','4.0','5.0','6.0']
        for cur_Rv in Rvs:
           ext_model = O94(Rv=cur_Rv)
           ax.plot(x,ext_model(x),label='R(V) = ' + str(cur_Rv))

        ax.set_xlabel(r'$x$ [$\mu m^{-1}$]')
        ax.set_ylabel(r'$A(x)/A(V)$')

        ax.legend(loc='best')
        plt.show()
    """

    Rv_range = [2.0, 6.0]
    x_range = x_range_O94

    @staticmethod
    def evaluate(in_x, Rv):
        """
        O94 function

        Parameters
        ----------
        in_x: float
           expects either x in units of wavelengths or frequency
           or assumes wavelengths in wavenumbers [1/micron]

           internally wavenumbers are used

        Returns
        -------
        axav: np array (float)
            A(x)/A(V) extinction curve [mag]

        Raises
        ------
        ValueError
           Input x values outside of defined range
        """
        x = _get_x_in_wavenumbers(in_x)

        # check that the wavenumbers are within the defined range
        _test_valid_x_range(x, x_range_O94, "O94")

        # setup the a & b coefficient vectors
        n_x = len(x)
        a = np.zeros(n_x)
        b = np.zeros(n_x)

        # define the ranges
        ir_indxs = np.where(np.logical_and(0.3 <= x, x < 1.1))
        opt_indxs = np.where(np.logical_and(1.1 <= x, x < 3.3))
        nuv_indxs = np.where(np.logical_and(3.3 <= x, x <= 8.0))
        fnuv_indxs = np.where(np.logical_and(5.9 <= x, x <= 8))
        fuv_indxs = np.where(np.logical_and(8 < x, x <= 10))

        # Infrared
        y = x[ir_indxs] ** 1.61
        a[ir_indxs] = 0.574 * y
        b[ir_indxs] = -0.527 * y

        # NIR/optical
        y = x[opt_indxs] - 1.82
        a[opt_indxs] = np.polyval(
            (-0.505, 1.647, -0.827, -1.718, 1.137, 0.701, -0.609, 0.104, 1), y
        )
        b[opt_indxs] = np.polyval(
            (3.347, -10.805, 5.491, 11.102, -7.985, -3.989, 2.908, 1.952, 0), y
        )

        # NUV
        a[nuv_indxs] = (
            1.752 - 0.316 * x[nuv_indxs] - 0.104 / ((x[nuv_indxs] - 4.67) ** 2 + 0.341)
        )
        b[nuv_indxs] = (
            -3.09 + 1.825 * x[nuv_indxs] + 1.206 / ((x[nuv_indxs] - 4.62) ** 2 + 0.263)
        )

        # far-NUV
        y = x[fnuv_indxs] - 5.9
        a[fnuv_indxs] += -0.04473 * (y**2) - 0.009779 * (y**3)
        b[fnuv_indxs] += 0.2130 * (y**2) + 0.1207 * (y**3)

        # FUV
        y = x[fuv_indxs] - 8.0
        a[fuv_indxs] = np.polyval((-0.070, 0.137, -0.628, -1.073), y)
        b[fuv_indxs] = np.polyval((0.374, -0.42, 4.257, 13.67), y)

        # return A(x)/A(V)
        return a + b / Rv


class F99(BaseExtRvModel):
    r"""
    Fitzpatrick (1999) Milky Way R(V) dependent model

    Parameters
    ----------
    Rv: float
        R(V) = A(V)/E(B-V) = total-to-selective extinction

    Raises
    ------
    InputParameterError
       Input Rv values outside of defined range

    Notes
    -----
    From Fitzpatrick (1999, PASP, 111, 63)

    Example showing F99 curves for a range of R(V) values.

    .. plot::
        :include-source:

        import numpy as np
        import matplotlib.pyplot as plt
        import astropy.units as u

        from dust_extinction.parameter_averages import F99

        fig, ax = plt.subplots()

        # temp model to get the correct x range
        text_model = F99()

        # generate the curves and plot them
        x = np.arange(text_model.x_range[0],
                      text_model.x_range[1],0.1)/u.micron

        Rvs = ['2.0','3.0','4.0','5.0','6.0']
        for cur_Rv in Rvs:
           ext_model = F99(Rv=cur_Rv)
           ax.plot(x,ext_model(x),label='R(V) = ' + str(cur_Rv))

        ax.set_xlabel(r'$x$ [$\mu m^{-1}$]')
        ax.set_ylabel(r'$A(x)/A(V)$')

        ax.legend(loc='best')
        plt.show()
    """

    Rv_range = [2.0, 6.0]
    x_range = x_range_F99

    def evaluate(self, in_x, Rv):
        """
        F99 function

        Parameters
        ----------
        in_x: float
           expects either x in units of wavelengths or frequency
           or assumes wavelengths in wavenumbers [1/micron]

           internally wavenumbers are used

        Returns
        -------
        axav: np array (float)
            A(x)/A(V) extinction curve [mag]

        Raises
        ------
        ValueError
           Input x values outside of defined range
        """
        # just in case someone calls evaluate explicitly
        Rv = np.atleast_1d(Rv)

        # ensure Rv is a single element, not numpy array
        Rv = Rv[0]

        # constant terms
        C3 = 3.23
        C4 = 0.41
        xo = 4.596
        gamma = 0.99

        # terms depending on Rv
        C2 = -0.824 + 4.717 / Rv
        # original F99 C1-C2 correlation
        C1 = 2.030 - 3.007 * C2

        # spline points
        optnir_axav_x = 10000.0 / np.array(
            [26500.0, 12200.0, 6000.0, 5470.0, 4670.0, 4110.0]
        )

        # determine optical/IR values at spline points
        #    Final optical spline point has a leading "-1.208" in Table 4
        #    of F99, but that does not reproduce Table 3.
        #    Additional indication that this is not correct is from
        #    fm_unred.pro
        #    which is based on FMRCURVE.pro distributed by Fitzpatrick.
        #    --> confirmation needed?
        #
        #    Also, fm_unred.pro has different coeff and # of terms,
        #    but later work does not include these terms
        #    --> check with Fitzpatrick?
        opt_axebv_y = np.array(
            [
                -0.426 + 1.0044 * Rv,
                -0.050 + 1.0016 * Rv,
                0.701 + 1.0016 * Rv,
                1.208 + 1.0032 * Rv - 0.00033 * (Rv**2),
            ]
        )
        nir_axebv_y = np.array([0.265, 0.829]) * Rv / 3.1
        optnir_axebv_y = np.concatenate([nir_axebv_y, opt_axebv_y])

        # return A(x)/A(V)
        return _curve_F99_method(
            in_x,
            Rv,
            C1,
            C2,
            C3,
            C4,
            xo,
            gamma,
            optnir_axav_x,
            optnir_axebv_y / Rv,
            self.x_range,
            self.__class__.__name__,
        )


class F04(BaseExtRvModel):
    r"""
    Fitzpatrick (2004) Milky Way R(V) dependent model

    Parameters
    ----------
    Rv: float
        R(V) = A(V)/E(B-V) = total-to-selective extinction

    Raises
    ------
    InputParameterError
       Input Rv values outside of defined range

    Notes
    -----
    From Fitzpatrick (2004, ASP Conf. Ser. 309, Astrophysics of Dust, 33)
        Equivalent to the F99 model with an updated NIR Rv dependence

    See also Fitzpatrick & Massa (2007, ApJ, 663, 320)

    Example showing F04 curves for a range of R(V) values.

    .. plot::
        :include-source:

        import numpy as np
        import matplotlib.pyplot as plt
        import astropy.units as u

        from dust_extinction.parameter_averages import F04

        fig, ax = plt.subplots()

        # temp model to get the correct x range
        text_model = F04()

        # generate the curves and plot them
        x = np.arange(text_model.x_range[0],
                      text_model.x_range[1],0.1)/u.micron

        Rvs = ['2.0','3.0','4.0','5.0','6.0']
        for cur_Rv in Rvs:
           ext_model = F04(Rv=cur_Rv)
           ax.plot(x,ext_model(x),label='R(V) = ' + str(cur_Rv))

        ax.set_xlabel(r'$x$ [$\mu m^{-1}$]')
        ax.set_ylabel(r'$A(x)/A(V)$')

        ax.legend(loc='best')
        plt.show()
    """

    Rv_range = [2.0, 6.0]
    x_range = x_range_F04

    def evaluate(self, in_x, Rv):
        """
        F04 function

        Parameters
        ----------
        in_x: float
           expects either x in units of wavelengths or frequency
           or assumes wavelengths in wavenumbers [1/micron]

           internally wavenumbers are used

        Returns
        -------
        axav: np array (float)
            A(x)/A(V) extinction curve [mag]

        Raises
        ------
        ValueError
           Input x values outside of defined range
        """
        # just in case someone calls evaluate explicitly
        Rv = np.atleast_1d(Rv)

        # ensure Rv is a single element, not numpy array
        Rv = Rv[0]

        # constant terms
        C3 = 2.991
        C4 = 0.319
        xo = 4.592
        gamma = 0.922

        # original F99 Rv dependence
        C2 = -0.824 + 4.717 / Rv
        # updated F04 C1-C2 correlation
        C1 = 2.18 - 2.91 * C2

        # spline points
        opt_axav_x = 10000.0 / np.array([6000.0, 5470.0, 4670.0, 4110.0])
        # **Use NIR spline x values in FM07, clipped to K band for now
        nir_axav_x = np.array([0.50, 0.75, 1.0])
        optnir_axav_x = np.concatenate([nir_axav_x, opt_axav_x])

        # **Keep optical spline points from F99:
        #    Final optical spline point has a leading "-1.208" in Table 4
        #    of F99, but that does not reproduce Table 3.
        #    Additional indication that this is not correct is from
        #    fm_unred.pro
        #    which is based on FMRCURVE.pro distributed by Fitzpatrick.
        #    --> confirmation needed?
        opt_axebv_y = np.array(
            [
                -0.426 + 1.0044 * Rv,
                -0.050 + 1.0016 * Rv,
                0.701 + 1.0016 * Rv,
                1.208 + 1.0032 * Rv - 0.00033 * (Rv**2),
            ]
        )
        # updated NIR curve from F04, note R dependence
        nir_axebv_y = (0.63 * Rv - 0.84) * nir_axav_x**1.84

        optnir_axebv_y = np.concatenate([nir_axebv_y, opt_axebv_y])

        # return A(x)/A(V)
        return _curve_F99_method(
            in_x,
            Rv,
            C1,
            C2,
            C3,
            C4,
            xo,
            gamma,
            optnir_axav_x,
            optnir_axebv_y / Rv,
            self.x_range,
            self.__class__.__name__,
        )


class VCG04(BaseExtRvModel):
    r"""
    Valencic, Clayton, & Gordon (2004) Milky Way R(V) dependent model

    Parameters
    ----------
    Rv: float
        R(V) = A(V)/E(B-V) = total-to-selective extinction

    Raises
    ------
    InputParameterError
       Input Rv values outside of defined range

    Notes
    -----
    From Valencic, Clayton, & Gordon (2004, ApJ, 616, 912)
    Including erratum: 2014, ApJ, 793, 66

    This model applies to the UV spectral region all the way to 912 A.
    This model was not derived for the optical or NIR.

    Example showing V04 curves for a range of R(V) values.

    .. plot::
        :include-source:

        import numpy as np
        import matplotlib.pyplot as plt
        import astropy.units as u

        from dust_extinction.parameter_averages import VCG04

        fig, ax = plt.subplots()

        # generate the curves and plot them
        x = np.arange(3.3,8.0, 0.1)/u.micron

        Rvs = ['2.0','3.0','4.0','5.0','6.0']
        for cur_Rv in Rvs:
           ext_model = VCG04(Rv=cur_Rv)
           ax.plot(x,ext_model(x),label='R(V) = ' + str(cur_Rv))

        ax.set_xlabel(r'$x$ [$\mu m^{-1}$]')
        ax.set_ylabel(r'$A(x)/A(V)$')

        ax.legend(loc='best')
        plt.show()
    """

    Rv_range = [2.0, 6.0]
    x_range = x_range_VCG04

    @staticmethod
    def evaluate(in_x, Rv):
        """
        VCG04 function

        Parameters
        ----------
        in_x: float
           expects either x in units of wavelengths or frequency
           or assumes wavelengths in wavenumbers [1/micron]
           internally wavenumbers are used

        Returns
        -------
        axav: np array (float)
            A(x)/A(V) extinction curve [mag]

        Raises
        ------
        ValueError
           Input x values outside of defined range
        """
        # convert to wavenumbers (1/micron) if x input in units
        # otherwise, assume x in appropriate wavenumber units
        x = _get_x_in_wavenumbers(in_x)

        # check that the wavenumbers are within the defined range
        _test_valid_x_range(x, x_range_VCG04, "VCG04")

        # setup the a & b coefficient vectors
        n_x = len(x)
        a = np.zeros(n_x)
        b = np.zeros(n_x)

        # define the ranges
        nuv_indxs = np.where(np.logical_and(3.3 <= x, x <= 8.0))
        fnuv_indxs = np.where(np.logical_and(5.9 <= x, x <= 8))

        # NUV
        a[nuv_indxs] = (
            1.808 - 0.215 * x[nuv_indxs] - 0.134 / ((x[nuv_indxs] - 4.558) ** 2 + 0.566)
        )
        b[nuv_indxs] = (
            -2.350
            + 1.403 * x[nuv_indxs]
            + 1.103 / ((x[nuv_indxs] - 4.587) ** 2 + 0.263)
        )

        # far-NUV
        y = x[fnuv_indxs] - 5.9
        a[fnuv_indxs] += -0.0077 * (y**2) - 0.0030 * (y**3)
        b[fnuv_indxs] += 0.2060 * (y**2) + 0.0550 * (y**3)

        # return A(x)/A(V)
        return a + b / Rv


class GCC09(BaseExtRvModel):
    r"""
    Grodon, Cartledge, & Clayton (2009) Milky Way R(V) dependent model

    Parameters
    ----------
    Rv: float
        R(V) = A(V)/E(B-V) = total-to-selective extinction

    Raises
    ------
    InputParameterError
       Input Rv values outside of defined range

    Notes
    -----
    From Gordon, Cartledge, & Clayton (2009, ApJ, 705, 1320)
    Including erratum: 2014, ApJ, 781, 128

    This model applies to the UV spectral region all the way to 912 A.
    This model was not derived for the optical or NIR.

    Example showing GCC09 curves for a range of R(V) values.

    .. plot::
        :include-source:

        import numpy as np
        import matplotlib.pyplot as plt
        import astropy.units as u

        from dust_extinction.parameter_averages import GCC09

        fig, ax = plt.subplots()

        # generate the curves and plot them
        x = np.arange(3.3, 11, 0.1)/u.micron

        Rvs = ['2.0','3.0','4.0','5.0','6.0']
        for cur_Rv in Rvs:
           ext_model = GCC09(Rv=cur_Rv)
           ax.plot(x,ext_model(x),label='R(V) = ' + str(cur_Rv))

        ax.set_xlabel(r'$x$ [$\mu m^{-1}$]')
        ax.set_ylabel(r'$A(x)/A(V)$')

        ax.legend(loc='best')
        plt.show()
    """

    Rv_range = [2.0, 6.0]
    x_range = x_range_GCC09

    @staticmethod
    def evaluate(in_x, Rv):
        """
        GCC09 function

        Parameters
        ----------
        in_x: float
           expects either x in units of wavelengths or frequency
           or assumes wavelengths in wavenumbers [1/micron]

           internally wavenumbers are used

        Returns
        -------
        axav: np array (float)
            A(x)/A(V) extinction curve [mag]

        Raises
        ------
        ValueError
           Input x values outside of defined range
        """
        # convert to wavenumbers (1/micron) if x input in units
        # otherwise, assume x in appropriate wavenumber units
        x = _get_x_in_wavenumbers(in_x)

        # check that the wavenumbers are within the defined range
        _test_valid_x_range(x, x_range_GCC09, "GCC09")

        # setup the a & b coefficient vectors
        n_x = len(x)
        a = np.zeros(n_x)
        b = np.zeros(n_x)

        # define the ranges
        nuv_indxs = np.where(np.logical_and(3.3 <= x, x <= 11.0))
        fnuv_indxs = np.where(np.logical_and(5.9 <= x, x <= 11.0))

        # NUV
        a[nuv_indxs] = (
            1.894
            - 0.373 * x[nuv_indxs]
            - 0.0101 / ((x[nuv_indxs] - 4.57) ** 2 + 0.0384)
        )
        b[nuv_indxs] = (
            -3.490 + 2.057 * x[nuv_indxs] + 0.706 / ((x[nuv_indxs] - 4.59) ** 2 + 0.169)
        )

        # far-NUV
        y = x[fnuv_indxs] - 5.9
        a[fnuv_indxs] += -0.110 * (y**2) - 0.0100 * (y**3)
        b[fnuv_indxs] += 0.531 * (y**2) + 0.0544 * (y**3)

        # return A(x)/A(V)
        return a + b / Rv


class M14(BaseExtRvModel):
    r"""
    Maiz Apellaniz et al (2014) Milky Way & LMC R(V) dependent model

    Parameters
    ----------
    R5495: float
        R5495 = A(5485)/E(4405-5495)
        Spectral equivalent to photometric R(V),
        standard value is 3.1

    Raises
    ------
    InputParameterError
       Input Rv values outside of defined range

    Notes
    -----
    M14 R5485-dependent model

    From Maiz Apellaniz et al. (2014, A&A, 564, 63),
    following structure of IDL code provided in paper appendix

    The published UV extinction curve is identical to Clayton, Cardelli,
    and Mathis (1989, CCM). Forcing the optical section to match
    smoothly with CCM introduces a non-physical feature at high
    values of R5495 around 3.9 inverse microns; see section 5 in
    Maiz Apellaniz et al. (2014) for more discussion.  For
    that reason, we provide the M14 model only through 3.3 inverse
    microns, the limit of the optical in CCM.

    R5495 = A(5485)/E(4405-5495)
    Spectral equivalent to photometric R(V),
    standard value is 3.1

    Example showing M14 curves for a range of R5495 values.

    .. plot::
        :include-source:

        import numpy as np
        import matplotlib.pyplot as plt
        import astropy.units as u

        from dust_extinction.parameter_averages import M14

        fig, ax = plt.subplots()

        # temp model to get the correct x range
        text_model = M14()

        # generate the curves and plot them
        x = np.arange(text_model.x_range[0],
                      text_model.x_range[1],0.1)/u.micron

        Rvs = ['2.0','3.1','4.0','5.0','6.0']
        for cur_Rv in Rvs:
           ext_model = M14(Rv=cur_Rv)
           ax.plot(x,ext_model(x),label='R(V) = ' + str(cur_Rv))

        ax.set_xlabel(r'$x$ [$\mu m^{-1}$]')
        ax.set_ylabel(r'$A(x)/A(V)$')

        ax.legend(loc='best')
        plt.show()
    """

    Rv_range = [2.0, 6.0]
    x_range = x_range_M14

    def evaluate(self, in_x, Rv):
        """
        M14 function

        Parameters
        ----------
        in_x: float
           expects either x in units of wavelengths or frequency
           or assumes wavelengths in wavenumbers [1/micron]

           internally wavenumbers are used

        Returns
        -------
        axav: np array (float)
            A(x)/A(V) extinction curve [mag]

        Raises
        ------
        ValueError
           Input x values outside of defined range
        """
        x = _get_x_in_wavenumbers(in_x)

        # check that the wavenumbers are within the defined range
        _test_valid_x_range(x, self.x_range, self.__class__.__name__)

        # just in case someone calls evaluate explicitly
        Rv = np.atleast_1d(Rv)

        # ensure Rv is a single element, not numpy array
        Rv = Rv[0]

        # Infrared
        ai = 0.574 * x**1.61
        bi = -0.527 * x**1.61

        # Optical
        x1 = np.array([1.0])
        xi1 = x1[0]
        x2 = np.array([1.15, 1.81984, 2.1, 2.27015, 2.7])
        x3 = np.array([3.5, 3.9, 4.0, 4.1, 4.2])
        xi3 = x3[-1]

        a1v = 0.574 * x1**1.61
        a1d = 0.574 * 1.61 * xi1**0.61
        b1v = -0.527 * x1**1.61
        b1d = -0.527 * 1.61 * xi1**0.61

        a2v = (
            1
            + 0.17699 * (x2 - 1.82)
            - 0.50447 * (x2 - 1.82) ** 2
            - 0.02427 * (x2 - 1.82) ** 3
            + 0.72085 * (x2 - 1.82) ** 4
            + 0.01979 * (x2 - 1.82) ** 5
            - 0.77530 * (x2 - 1.82) ** 6
            + 0.32999 * (x2 - 1.82) ** 7
            + np.array([0.0, 0.0, -0.011, 0.0, 0.0])
        )
        b2v = (
            1.41338 * (x2 - 1.82)
            + 2.28305 * (x2 - 1.82) ** 2
            + 1.07233 * (x2 - 1.82) ** 3
            - 5.38434 * (x2 - 1.82) ** 4
            - 0.62251 * (x2 - 1.82) ** 5
            + 5.30260 * (x2 - 1.82) ** 6
            - 2.09002 * (x2 - 1.82) ** 7
            + np.array([0.0, 0.0, +0.091, 0.0, 0.0])
        )

        a3v = (
            1.752
            - 0.316 * x3
            - 0.104 / ((x3 - 4.67) ** 2 + 0.341)
            + np.array([0.442, 0.341, 0.130, 0.020, 0.000])
        )
        a3d = -0.316 + 0.104 * 2.0 * (xi3 - 4.67) / ((xi3 - 4.67) ** 2 + 0.341) ** 2
        b3v = (
            -3.090
            + 1.825 * x3
            + 1.206 / ((x3 - 4.62) ** 2 + 0.263)
            - np.array([1.256, 1.021, 0.416, 0.064, 0.000])
        )
        b3d = 1.825 - 1.206 * 2 * (xi3 - 4.62) / ((xi3 - 4.62) ** 2 + 0.263) ** 2

        xn = np.concatenate((x1, x2, x3))
        anv = np.concatenate((a1v, a2v, a3v))
        bnv = np.concatenate((b1v, b2v, b3v))

        a_spl = interpolate.CubicSpline(xn, anv, bc_type=((1, a1d), (1, a3d)))
        b_spl = interpolate.CubicSpline(xn, bnv, bc_type=((1, b1d), (1, b3d)))

        av = a_spl(x)
        bv = b_spl(x)

        # UV extinction curve in the paper repeats CCM. Forcing the
        # optical section to match smoothly with CCM introduces a
        # non-physical feature at high values of R5495 at x = 3.9
        # inverse microns.  This class does not provide the UV curve,
        # but the code that would calculate it is included below for
        # completeness.

        # Ultraviolet
        # y = x - 5.9
        # fa = np.zeros(x.size)
        #      + (-0.04473*y**2 - 0.009779*y**3)*((x<8)&(x>5.9))
        # fb = np.zeros(x.size) + ( 0.2130*y**2 + 0.1207*y**3)*((x<8)&(x>5.9))

        # au = 1.752 - 0.316*x - 0.104/((x-4.67)**2 + 0.341) + fa
        # bu = -3.090 + 1.825*x + 1.206/((x-4.62)**2 + 0.263) + fb

        # Far ultraviolet
        # y = x - 8.0
        # af = -1.073 - 0.628*y + 0.137*y**2 - 0.070*y**3
        # bf = 13.670 + 4.257*y - 0.420*y**2 + 0.374*y**3

        # Final result
        # a = (ai*(x<xi1) + av*((x>xi1) & (x<xi3))
        #         + au*((x>xi3)&(x<8.0)) + af*(x>8.0))
        # b = (bi*(x<xi1) + bv*((x>xi1) & (x<xi3))
        #         + bu*((x>xi3)&(x<8.0)) + bf*(x>8.0))

        # Final result
        a = ai * (x < xi1) + av * ((x >= xi1) & (x < xi3))
        b = bi * (x < xi1) + bv * ((x >= xi1) & (x < xi3))

        return a + b / Rv


class G16(BaseExtRvAfAModel):
    r"""
    Gordon et al (2016) Milky Way, LMC, & SMC R(V) and f_A dependent model

    Mixture model between the F99 R(V) dependent model (component A)
    and the G03_SMCBar model (component B)

    Parameters
    ----------
    RvA: float
         R_A(V) = A(V)/E(B-V) = total-to-selective extinction
         R(V) of the A component

    fA: float
        f_A is the mixture coefficent between the R(V)

    Raises
    ------
    InputParameterError
       Input RvA values outside of defined range
       Input fA values outside of defined range

    Notes
    -----
    From Gordon et al. (2016, ApJ, 826, 104)

    Example showing G16 curves for a range of R_A(V) values
    and f_A values.

    .. plot::
        :include-source:

        import numpy as np
        import matplotlib.pyplot as plt
        import astropy.units as u

        from dust_extinction.parameter_averages import G16

        fig, ax = plt.subplots()

        # temp model to get the correct x range
        text_model = G16()

        # generate the curves and plot them
        x = np.arange(text_model.x_range[0],
                      text_model.x_range[1],0.1)/u.micron

        Rvs = ['2.0','3.0','4.0','5.0','6.0']
        for cur_Rv in Rvs:
           ext_model = G16(RvA=cur_Rv, fA=1.0)
           ax.plot(x,ext_model(x),label=r'$R_A(V) = ' + str(cur_Rv) + '$')

        ax.set_xlabel(r'$x$ [$\mu m^{-1}$]')
        ax.set_ylabel(r'$A(x)/A(V)$')

        ax.legend(loc='best', title=r'$f_A = 1.0$')
        plt.show()

    .. plot::
        :include-source:

        import numpy as np
        import matplotlib.pyplot as plt
        import astropy.units as u

        from dust_extinction.parameter_averages import G16

        fig, ax = plt.subplots()

        # temp model to get the correct x range
        text_model = G16()

        # generate the curves and plot them
        x = np.arange(text_model.x_range[0],
                      text_model.x_range[1],0.1)/u.micron

        fAs = [0.0, 0.2, 0.4, 0.6, 0.8, 1.0]
        for cur_fA in fAs:
           ext_model = G16(RvA=3.1, fA=cur_fA)
           ax.plot(x,ext_model(x),label=r'$f_A = ' + str(cur_fA) + '$')

        ax.set_xlabel(r'$x$ [$\mu m^{-1}$]')
        ax.set_ylabel(r'$A(x)/A(V)$')

        ax.legend(loc='best', title=r'$R_A(V) = 3.1$')
        plt.show()
    """

    RvA_range = [2.0, 6.0]
    fA_range = [0.0, 1.0]
    x_range = x_range_G16

    @staticmethod
    def evaluate(in_x, RvA, fA):
        """
        G16 function

        Parameters
        ----------
        in_x: float
           expects either x in units of wavelengths or frequency
           or assumes wavelengths in wavenumbers [1/micron]

           internally wavenumbers are used

        Returns
        -------
        axav: np array (float)
            A(x)/A(V) extinction curve [mag]

        Raises
        ------
        ValueError
           Input x values outside of defined range
        """
        x = _get_x_in_wavenumbers(in_x)

        # check that the wavenumbers are within the defined range
        _test_valid_x_range(x, x_range_G16, "G16")

        # just in case someone calls evaluate explicitly
        RvA = np.atleast_1d(RvA)

        # ensure Rv is a single element, not numpy array
        RvA = RvA[0]

        # get the A component extinction model
        extA_model = F99(Rv=RvA)
        alav_A = extA_model(x / u.micron)

        # get the B component extinction model
        extB_model = G03_SMCBar()
        alav_B = extB_model(x / u.micron)

        # create the mixture model
        alav = fA * alav_A + (1.0 - fA) * alav_B

        # return A(x)/A(V)
        return alav


class F19(BaseExtRvModel):
    r"""
    Fitzpatrick et al (2019) extinction model calculation

    Fitzpatrick, Massa, Gordon et al. (2019, ApJ, 886, 108) model.
    Based on a sample of stars observed spectroscopically in the
    optical with HST/STIS.

    Parameters
    ----------
    Rv: float
        R(V) = A(V)/E(B-V) = total-to-selective extinction

    Raises
    ------
    InputParameterError
       Input Rv values outside of defined range

    Notes
    -----
    F19 Milky Way R(V) dependent extinction model

    Example showing F19 curves for a range of R(V) values.

    .. plot::
        :include-source:

        import numpy as np
        import matplotlib.pyplot as plt
        import astropy.units as u

        from dust_extinction.parameter_averages import F19

        fig, ax = plt.subplots()

        # temp model to get the correct x range
        text_model = F19()

        # generate the curves and plot them
        x = np.arange(text_model.x_range[0],
                      text_model.x_range[1],0.1)/u.micron

        Rvs = [2.0, 3.0, 4.0, 5.0, 6.0]
        for cur_Rv in Rvs:
           ext_model = F19(Rv=cur_Rv)
           ax.plot(x,ext_model(x),label='R(V) = ' + str(cur_Rv))

        ax.set_xlabel(r'$x$ [$\mu m^{-1}$]')
        ax.set_ylabel(r'$A(x)/A(V)$')

        ax.legend(loc='best')
        plt.show()
    """

    Rv_range = [2.0, 6.0]
    x_range = [0.3, 8.7]

    def __init__(self, Rv=3.1, **kwargs):

        # get the tabulated information
        data_path = pkg_resources.resource_filename("dust_extinction", "data/")

        a = Table.read(data_path + "F19_tabulated.dat", format="ascii")

        # compute E(lambda-55)/E(B-55) on the tabulated x points
        self.k_rV_tab_x = a["k_3.02"].data + a["deltak"].data * (Rv - 3.10) * 0.990

        # setup spline interpolation
        self.spline_rep = interpolate.splrep(a["x"].data, self.k_rV_tab_x)

        super().__init__(Rv, **kwargs)

    def evaluate(self, in_x, Rv):
        """
        F19 function

        Parameters
        ----------
        in_x: float
           expects either x in units of wavelengths or frequency
           or assumes wavelengths in wavenumbers [1/micron]

           internally wavenumbers are used

        Returns
        -------
        axav: np array (float)
            A(x)/A(V) extinction curve [mag]

        Raises
        ------
        ValueError
           Input x values outside of defined range
        """
        # convert to wavenumbers (1/micron) if x input in units
        # otherwise, assume x in appropriate wavenumber units
        x = _get_x_in_wavenumbers(in_x)

        # check that the wavenumbers are within the defined range
        _test_valid_x_range(x, self.x_range, self.__class__.__name__)

        # just in case someone calls evaluate explicitly
        Rv = np.atleast_1d(Rv)

        # ensure Rv is a single element, not numpy array
        Rv = Rv[0]

        # use spline interpolation to evaluate the curve for the input x values
        k_rV = interpolate.splev(x, self.spline_rep, der=0)

        # convert to A(x)/A(55) from E(x-55)/E(44-55)
        a_rV = k_rV / Rv + 1.0

        # return A(x)/A(55)
        return a_rV


class D22(BaseExtRvModel):
    r"""
    Decleir et al (2022) extinction model calculation

    Decleir, Gordon, et al. (2022, ApJ, submitted) model.
    Based on a sample of stars observed spectroscopically in the
    NIR with IRTF/SpeX.

    Parameters
    ----------
    Rv: float
        R(V) = A(V)/E(B-V) = total-to-selective extinction

    Raises
    ------
    InputParameterError
       Input Rv values outside of defined range

    Notes
    -----
    D22 Milky Way R(V) dependent extinction model

    Example showing D22 curves for a range of R(V) values.

    .. plot::
        :include-source:

        import numpy as np
        import matplotlib.pyplot as plt
        import astropy.units as u

        from dust_extinction.parameter_averages import D22

        fig, ax = plt.subplots()

        # temp model to get the correct x range
        text_model = D22()

        # generate the curves and plot them
        x = np.arange(text_model.x_range[0], text_model.x_range[1], 0.01) / u.micron

        Rvs = [2.5, 3.1, 4.0, 4.75, 5.5]
        for cur_Rv in Rvs:
            ext_model = D22(Rv=cur_Rv)
            ax.plot(1. / x, ext_model(x), label="R(V) = " + str(cur_Rv))

        ax.set_xlabel(r"$\lambda$ [$\mu m$]")
        ax.set_ylabel(r"$A(x)/A(V)$")

        ax.legend(loc="best")
        plt.show()
    """

    Rv_range = [2.5, 5.5]
    x_range = [1 / 4.0, 1 / 0.80]

    def __init__(self, Rv=3.1, **kwargs):

        # get the tabulated information
        data_path = pkg_resources.resource_filename("dust_extinction", "data/")

        a = Table.read(data_path + "D22_Rv_slope.dat", format="ascii")

        # setup spline interpolation
        self.spline_rep = interpolate.splrep(a["wavelength[micron]"].data, a["slope"])

        super().__init__(Rv, **kwargs)

    def evaluate(self, in_x, Rv):
        """
        D22 function

        Parameters
        ----------
        in_x: float
           expects either x in units of wavelengths or frequency
           or assumes wavelengths in wavenumbers [1/micron]

           internally wavenumbers are used

        Returns
        -------
        axav: np array (float)
            A(x)/A(V) extinction curve [mag]

        Raises
        ------
        ValueError
           Input x values outside of defined range
        """
        # convert to wavenumbers (1/micron) if x input in units
        # otherwise, assume x in appropriate wavenumber units
        x = _get_x_in_wavenumbers(in_x)

        # check that the wavenumbers are within the defined range
        _test_valid_x_range(x, self.x_range, self.__class__.__name__)

        # just in case someone calls evaluate explicitly
        Rv = np.atleast_1d(Rv)

        # ensure Rv is a single element, not numpy array
        Rv = Rv[0]

        # intercepts
        mod_a = PowerLaw1D(amplitude=0.377, alpha=1.78, x_0=1.0)
        a = mod_a(1.0 / x)

        # slopes
        # from spline interpolation
        b = interpolate.splev(1.0 / x, self.spline_rep, der=0)

        # return A(x)/A(V)
        return a + b * (1.0 / Rv - 1 / 3.1)


class G23(BaseExtRvModel):
    r"""
    Gordon et al. (2023) Milky Way R(V) dependent model

    Parameters
    ----------
    Rv: float
        R(V) = A(V)/E(B-V) = total-to-selective extinction

    Raises
    ------
    InputParameterError
       Input Rv values outside of defined range

    Notes
    -----
    From Gordon et al. (2023, ApJ, in press)

    Example showing G23 curves for a range of R(V) values.

    .. plot::
        :include-source:

        import numpy as np
        import matplotlib.pyplot as plt
        import astropy.units as u

        from dust_extinction.parameter_averages import G23

        fig, ax = plt.subplots()

        # generate the curves and plot them
        lam = np.logspace(np.log10(0.0912), np.log10(30.0), num=1000) * u.micron

        Rvs = [2.5, 3.1, 4.0, 4.75, 5.5]
        for cur_Rv in Rvs:
           ext_model = G23(Rv=cur_Rv)
           ax.plot(lam,ext_model(lam),label='R(V) = ' + str(cur_Rv))

        ax.set_xscale('log')
        ax.set_yscale('log')

        ax.set_xlabel('$\lambda$ [$\mu$m]')
        ax.set_ylabel(r'$A(x)/A(V)$')

        ax.legend(loc='best')
        plt.show()
    """

    Rv_range = [2.3, 5.6]
    x_range = x_range_G23

    def evaluate(self, in_x, Rv):
        """
        G23 function

        Parameters
        ----------
        in_x: float
           expects either x in units of wavelengths or frequency
           or assumes wavelengths in wavenumbers [1/micron]

           internally wavenumbers are used

        Returns
        -------
        axav: np array (float)
            A(x)/A(V) extinction curve [mag]

        Raises
        ------
        ValueError
           Input x values outside of defined range
        """
        x = _get_x_in_wavenumbers(in_x)

        # check that the wavenumbers are within the defined range
        _test_valid_x_range(x, self.x_range, "G23")

        # setup the a & b coefficient vectors
        n_x = len(x)
        self.a = np.zeros(n_x)
        self.b = np.zeros(n_x)

        # define the ranges
        ir_indxs = np.where(np.logical_and(1.0 / 35.0 <= x, x < 1.0 / 1.0))
        opt_indxs = np.where(np.logical_and(1.0 / 1.1 <= x, x < 1.0 / 0.3))
        uv_indxs = np.where(np.logical_and(1.0 / 0.3 <= x, x <= 1.0 / 0.09))

        # overlap ranges
        optir_waves = [0.9, 1.1]
        optir_overlap = (x >= 1.0 / optir_waves[1]) & (x <= 1.0 / optir_waves[0])
        uvopt_waves = [0.3, 0.33]
        uvopt_overlap = (x >= 1.0 / uvopt_waves[1]) & (x <= 1.0 / uvopt_waves[0])

        # NIR/MIR
        # fmt: off
        # (scale, alpha1, alpha2, swave, swidth), sil1, sil2
        ir_a = [0.38526, 1.68467, 0.78791, 4.30578, 4.78338,
                0.06652, 9.8434, 2.21205, -0.24703,
                0.0267 , 19.58294, 17., -0.27]
        # fmt: on
        ir_b = [-1.01251, 1.0, -1.06099]
        self.a[ir_indxs] = self.nirmir_intercept(x[ir_indxs], ir_a)

        irpow = PowerLaw1D()
        irpow.parameters = ir_b
        self.b[ir_indxs] = irpow(x[ir_indxs])

        # optical
        # fmt: off
        # polynomial coeffs, ISS1, ISS2, ISS3
        opt_a = [-0.35848, 0.7122 , 0.08746, -0.05403, 0.00674,
                 0.03893, 2.288, 0.243,
                 0.02965, 2.054, 0.179,
                 0.01747, 1.587, 0.243]
        opt_b = [0.12354, -2.68335, 2.01901, -0.39299, 0.03355,
                 0.18453, 2.288, 0.243,
                 0.19728, 2.054, 0.179,
                 0.1713 , 1.587, 0.243]
        # fmt: on
        m20_model_a = Polynomial1D(4) + Drude1D() + Drude1D() + Drude1D()
        m20_model_a.parameters = opt_a
        self.a[opt_indxs] = m20_model_a(x[opt_indxs])
        m20_model_b = Polynomial1D(4) + Drude1D() + Drude1D() + Drude1D()
        m20_model_b.parameters = opt_b
        self.b[opt_indxs] = m20_model_b(x[opt_indxs])

        # overlap between optical/ir
        # weights = (1.0 / optir_waves[1] - x[optir_overlap]) / (
        #     1.0 / optir_waves[1] - 1.0 / optir_waves[0]
        # )
        weights = _smoothstep(
            1.0 / x[optir_overlap], x_min=optir_waves[0], x_max=optir_waves[1], N=1
        )
        self.a[optir_overlap] = (1.0 - weights) * m20_model_a(x[optir_overlap])
        self.a[optir_overlap] += weights * self.nirmir_intercept(x[optir_overlap], ir_a)
        self.b[optir_overlap] = (1.0 - weights) * m20_model_b(x[optir_overlap])
        self.b[optir_overlap] += weights * irpow(x[optir_overlap])

        # Ultraviolet
        uv_a = [0.81297, 0.2775, 1.06295, 0.11303, 4.60, 0.99]
        uv_b = [-2.97868, 1.89808, 3.10334, 0.65484, 4.60, 0.99]

        fm90_model_a = FM90()
        fm90_model_a.parameters = uv_a
        self.a[uv_indxs] = fm90_model_a(x[uv_indxs] / u.micron)
        fm90_model_b = FM90()
        fm90_model_b.parameters = uv_b
        self.b[uv_indxs] = fm90_model_b(x[uv_indxs] / u.micron)

        # overlap between uv/optical
        # weights = (1.0 / uvopt_waves[1] - x[uvopt_overlap]) / (
        #     1.0 / uvopt_waves[1] - 1.0 / uvopt_waves[0]
        # )
        weights = _smoothstep(
            1.0 / x[uvopt_overlap], x_min=uvopt_waves[0], x_max=uvopt_waves[1], N=1
        )
        self.a[uvopt_overlap] = (1.0 - weights) * fm90_model_a(
            x[uvopt_overlap] / u.micron
        )
        self.a[uvopt_overlap] += weights * m20_model_a(x[uvopt_overlap])
        self.b[uvopt_overlap] = (1.0 - weights) * fm90_model_b(
            x[uvopt_overlap] / u.micron
        )
        self.b[uvopt_overlap] += weights * m20_model_b(x[uvopt_overlap])

        # return A(x)/A(V)
        return self.a + self.b * (1 / Rv - 1 / 3.1)

    @staticmethod
    def nirmir_intercept(x, params):
        """
        Functional form for the NIR/MIR intercept term.
        Based on modifying the G21 shape model to have two power laws instead
        of one with a break wavelength.

        Parameters
        ----------
        x: float
           expects x in wavenumbers [1/micron]
        params: floats
           paramters of function

        Returns
        -------
        axav: np array (float)
            A(x)/A(V) extinction curve [mag]
        """
        wave = 1 / x

        # fmt: off
        (scale, alpha, alpha2, swave, swidth,
            sil1_amp, sil1_center, sil1_fwhm, sil1_asym,
            sil2_amp, sil2_center, sil2_fwhm, sil2_asym) = params
        # fmt: on

        # broken powerlaw with a smooth transition
        axav_pow1 = scale * (wave ** (-1.0 * alpha))

        norm_ratio = swave ** (-1.0 * alpha) / swave ** (-1.0 * alpha2)
        axav_pow2 = scale * norm_ratio * (wave ** (-1.0 * alpha2))

        # use smoothstep to smoothly transition between the two powerlaws
        weights = _smoothstep(
            wave, x_min=swave - swidth / 2, x_max=swave + swidth / 2, N=1
        )
        axav = axav_pow1 * (1.0 - weights) + axav_pow2 * weights

        # silicate feature drudes
        axav += _modified_drude(wave, sil1_amp, sil1_center, sil1_fwhm, sil1_asym)
        axav += _modified_drude(wave, sil2_amp, sil2_center, sil2_fwhm, sil2_asym)

        return axav



import numpy as np
from scipy import interpolate

import astropy.units as u
from astropy.modeling import Fittable1DModel, Parameter

from .helpers import _get_x_in_wavenumbers, _test_valid_x_range

__all__ = ["FM90", "FM90_B3", "P92", "G21"]

x_range_FM90 = [1.0 / 0.35, 1.0 / 0.09]
x_range_P92 = [1.0 / 1e3, 1.0 / 1e-3]


def _curve_F99_method(
    in_x,
    Rv,
    C1,
    C2,
    C3,
    C4,
    xo,
    gamma,
    optnir_axav_x,
    optnir_axav_y,
    valid_x_range,
    model_name,
):
    """
    Function to return extinction using F99 method

    Parameters
    ----------
    in_x: float
        expects either x in units of wavelengths or frequency
        or assumes wavelengths in wavenumbers [1/micron]

        internally wavenumbers are used

    Rv: float
       ratio of total to selective extinction = A(V)/E(B-V)

    C1: float
        y-intercept of linear term: FM90 parameter

    C2: float
        slope of liner term: FM90 parameter

    C3: float
        amplitude of "2175 A" bump: FM90 parameter

    C4: float
        amplitude of FUV rise: FM90 parameter

    xo: float
        centroid of "2175 A" bump: FM90 parameter

    gamma: float
        width of "2175 A" bump: FM90 parameter

    optnir_axav_x: float array
        vector of x values for optical/NIR A(x)/A(V) curve

    optnir_axav_y: float array
        vector of y values for optical/NIR A(x)/A(V) curve

    Returns
    -------
    axav: np array (float)
        A(x)/A(V) extinction curve [mag]

    Raises
    ------
    ValueError
        Input x values outside of defined range
    """
    x = _get_x_in_wavenumbers(in_x)

    # check that the wavenumbers are within the defined range
    _test_valid_x_range(x, valid_x_range, model_name)

    # initialize extinction curve storage
    axav = np.zeros(len(x))

    # x value above which FM90 parametrization used
    x_cutval_uv = 10000.0 / 2700.0

    # required UV points for spline interpolation
    x_splineval_uv = 10000.0 / np.array([2700.0, 2600.0])

    # UV points in input x
    (indxs_uv,) = np.where(x >= x_cutval_uv)

    # add in required spline points, otherwise just spline points
    if len(indxs_uv) > 0:
        xuv = np.concatenate([x_splineval_uv, x[indxs_uv]])
    else:
        xuv = x_splineval_uv

    # FM90 model and values
    fm90_model = FM90(C1=C1, C2=C2, C3=C3, C4=C4, xo=xo, gamma=gamma)
    # evaluate model and get results in A(x)/A(V)
    axav_fm90 = fm90_model(xuv / u.micron) / Rv + 1.0

    # save spline points
    y_splineval_uv = axav_fm90[0:2]

    # ingore the spline points
    if len(indxs_uv) > 0:
        axav[indxs_uv] = axav_fm90[2:]

    # **Optical Portion**
    #   using cubic spline anchored in UV, optical, and IR

    # optical/NIR points in input x
    (indxs_opir,) = np.where(x < x_cutval_uv)

    if len(indxs_opir) > 0:
        # spline points
        x_splineval_optir = optnir_axav_x

        # determine optical/IR values at spline points
        y_splineval_optir = optnir_axav_y

        # add in zero extinction at infinite wavelength
        x_splineval_optir = np.insert(x_splineval_optir, 0, 0.0)
        y_splineval_optir = np.insert(y_splineval_optir, 0, 0.0)

        spline_x = np.concatenate([x_splineval_optir, x_splineval_uv])
        spline_y = np.concatenate([y_splineval_optir, y_splineval_uv])
        spline_rep = interpolate.splrep(spline_x, spline_y)
        axav[indxs_opir] = interpolate.splev(x[indxs_opir], spline_rep, der=0)

    # return A(x)/A(V)
    return axav


def _modified_drude(x, scale, x_o, gamma_o, asym):
    """
    Modified Drude function to have a variable asymmetry.  Drude profiles
    are intrinsically asymmetric with the asymmetry fixed by specific central
    wavelength and width.  This modified Drude introduces an asymmetry
    parameter that allows for variable asymmetry at fixed central wavelength
    and width.

    Parameters
    ----------
    x : float
        input wavelengths

    scale : float
        central amplitude

    x_o : float
        central wavelength

    gamma_o : float
        full-width-half-maximum of profile

    asym : float
        asymmetry where a value of 0 results in a standard Drude profile
    """
    gamma = 2.0 * gamma_o / (1.0 + np.exp(asym * (x - x_o)))
    y = scale * ((gamma / x_o) ** 2) / ((x / x_o - x_o / x) ** 2 + (gamma / x_o) ** 2)

    return y


class FM90(Fittable1DModel):
    r"""
    Fitzpatrick & Massa (1990) 6 parameter ultraviolet shape model

    Parameters
    ----------
    C1: float
       y-intercept of linear term

    C2: float
       slope of liner term

    C3: float
       strength of "2175 A" bump (true amplitude is C3/gamma^2)

    C4: float
       amplitude of FUV rise

    xo: float
       centroid of "2175 A" bump

    gamma: float
       width of "2175 A" bump

    Notes
    -----
    From Fitzpatrick & Massa (1990, ApJS, 72, 163)

    Only applicable at UV wavelengths

    Example showing a FM90 curve with components identified.

    .. plot::
        :include-source:

        import numpy as np
        import matplotlib.pyplot as plt
        import astropy.units as u

        from dust_extinction.shapes import FM90

        fig, ax = plt.subplots()

        # generate the curves and plot them
        x = np.arange(3.8,8.6,0.1)/u.micron

        ext_model = FM90()
        ax.plot(x,ext_model(x),label='total')

        ext_model = FM90(C3=0.0, C4=0.0)
        ax.plot(x,ext_model(x),label='linear term')

        ext_model = FM90(C1=0.0, C2=0.0, C4=0.0)
        ax.plot(x,ext_model(x),label='bump term')

        ext_model = FM90(C1=0.0, C2=0.0, C3=0.0)
        ax.plot(x,ext_model(x),label='FUV rise term')

        ax.set_xlabel(r'$x$ [$\mu m^{-1}$]')
        ax.set_ylabel(r'$E(\lambda - V)/E(B - V)$')

        ax.legend(loc='best')
        plt.show()
    """
    n_inputs = 1
    n_outputs = 1

    C1 = Parameter(description="linear term: y-intercept", default=0.10)
    C2 = Parameter(description="linear term: slope", default=0.70)
    C3 = Parameter(description="bump: amplitude", default=3.23)
    C4 = Parameter(description="FUV rise: amplitude", default=0.41)
    xo = Parameter(description="bump: centroid", default=4.60, min=0.0)
    gamma = Parameter(description="bump: width", default=0.99, min=0.0)

    x_range = x_range_FM90

    @staticmethod
    def evaluate(in_x, C1, C2, C3, C4, xo, gamma):
        """
        FM90 function

        Parameters
        ----------
        in_x: float
           expects either x in units of wavelengths or frequency
           or assumes wavelengths in wavenumbers [1/micron]

           internally wavenumbers are used

        Returns
        -------
        exvebv: np array (float)
            E(x-V)/E(B-V) extinction curve [mag]

        Raises
        ------
        ValueError
           Input x values outside of defined range
        """
        x = _get_x_in_wavenumbers(in_x)

        # check that the wavenumbers are within the defined range
        _test_valid_x_range(x, x_range_FM90, "FM90")

        # linear term
        exvebv = C1 + C2 * x

        # bump term
        x2 = x**2
        exvebv += C3 * (x2 / ((x2 - xo**2) ** 2 + x2 * (gamma**2)))

        # FUV rise term
        fnuv_indxs = np.where(x >= 5.9)
        if len(fnuv_indxs) > 0:
            y = x[fnuv_indxs] - 5.9
            exvebv[fnuv_indxs] += C4 * (0.5392 * (y**2) + 0.05644 * (y**3))

        # return E(x-V)/E(B-V)
        return exvebv

    @staticmethod
    def fit_deriv(in_x, C1, C2, C3, C4, xo, gamma):
        """
        Derivatives of the FM90 function with respect to the parameters
        """
        x = in_x

        # useful quantitites
        x2 = x**2
        xo2 = xo**2
        g2 = gamma**2
        x2mxo2_2 = (x2 - xo2) ** 2
        denom = (x2mxo2_2 - x2 * g2) ** 2

        # derivatives
        d_C1 = np.full((len(x)), 1.0)
        d_C2 = x

        d_C3 = x2 / (x2mxo2_2 + x2 * g2)

        d_xo = (4.0 * C2 * x2 * xo * (x2 - xo2)) / denom

        d_gamma = (2.0 * C2 * (x2**2) * gamma) / denom

        d_C4 = np.zeros((len(x)))
        fuv_indxs = np.where(x >= 5.9)
        if len(fuv_indxs) > 0:
            y = x[fuv_indxs] - 5.9
            d_C4[fuv_indxs] = 0.5392 * (y**2) + 0.05644 * (y**3)

        return [d_C1, d_C2, d_C3, d_C4, d_xo, d_gamma]

    # @property
    # def input_units(self):
    #     if self.xo.unit is None:
    #         return None
    #     return {self.inputs[0]: self.xo.unit}
    #
    # def _parameter_units_for_data_units(self, inputs_unit, outputs_unit):
    #     return {'xo': inputs_unit[self.inputs[0]],
    #             'C1': outputs_unit[self.outputs[0]],
    #             'C2': outputs_unit[self.outputs[0]],
    #             'C3': outputs_unit[self.outputs[0]],
    #             'C4': outputs_unit[self.outputs[0]]}


class FM90_B3(Fittable1DModel):
    r"""
    Fitzpatrick & Massa (1990) 6 parameter ultraviolet shape model
    Version with bump amplitude B3 = C3/gamma^2

    Parameters
    ----------
    C1: float
       y-intercept of linear term

    C2: float
       slope of liner term

    B3: float
       amplitude of "2175 A" bump

    C4: float
       amplitude of FUV rise

    xo: float
       centroid of "2175 A" bump

    gamma: float
       width of "2175 A" bump

    Notes
    -----
    From Fitzpatrick & Massa (1990, ApJS, 72, 163)

    Only applicable at UV wavelengths

    Example showing a FM90 curve with components identified.

    .. plot::
        :include-source:

        import numpy as np
        import matplotlib.pyplot as plt
        import astropy.units as u

        from dust_extinction.shapes import FM90_B3

        fig, ax = plt.subplots()

        # generate the curves and plot them
        x = np.arange(3.8,8.6,0.1)/u.micron

        ext_model = FM90_B3()
        ax.plot(x,ext_model(x),label='total')

        ext_model = FM90_B3(B3=0.0, C4=0.0)
        ax.plot(x,ext_model(x),label='linear term')

        ext_model = FM90_B3(C1=0.0, C2=0.0, C4=0.0)
        ax.plot(x,ext_model(x),label='bump term')

        ext_model = FM90_B3(C1=0.0, C2=0.0, B3=0.0)
        ax.plot(x,ext_model(x),label='FUV rise term')

        ax.set_xlabel(r'$x$ [$\mu m^{-1}$]')
        ax.set_ylabel(r'$E(\lambda - V)/E(B - V)$')

        ax.legend(loc='best')
        plt.show()
    """
    n_inputs = 1
    n_outputs = 1

    C1 = Parameter(description="linear term: y-intercept", default=0.10)
    C2 = Parameter(description="linear term: slope", default=0.70)
    B3 = Parameter(description="bump: amplitude", default=3.23)
    C4 = Parameter(description="FUV rise: amplitude", default=0.41)
    xo = Parameter(description="bump: centroid", default=4.60, min=0.0)
    gamma = Parameter(description="bump: width", default=0.99, min=0.0)

    x_range = x_range_FM90

    @staticmethod
    def evaluate(in_x, C1, C2, B3, C4, xo, gamma):
        """
        FM90 function

        Parameters
        ----------
        in_x: float
           expects either x in units of wavelengths or frequency
           or assumes wavelengths in wavenumbers [1/micron]

           internally wavenumbers are used

        Returns
        -------
        exvebv: np array (float)
            E(x-V)/E(B-V) extinction curve [mag]

        Raises
        ------
        ValueError
           Input x values outside of defined range
        """
        x = _get_x_in_wavenumbers(in_x)

        # check that the wavenumbers are within the defined range
        _test_valid_x_range(x, x_range_FM90, "FM90_B3")

        # linear term
        exvebv = C1 + C2 * x

        # bump term
        x2 = x**2
        exvebv += (B3 * gamma**2) * (x2 / ((x2 - xo**2) ** 2 + x2 * (gamma**2)))

        # FUV rise term
        fnuv_indxs = np.where(x >= 5.9)
        if len(fnuv_indxs) > 0:
            y = x[fnuv_indxs] - 5.9
            exvebv[fnuv_indxs] += C4 * (0.5392 * (y**2) + 0.05644 * (y**3))

        # return E(x-V)/E(B-V)
        return exvebv


class P92(Fittable1DModel):
    r"""
    Pei (1992) 24 parameter shape model

    Parameters
    ----------
    BKG_amp : float
      background term amplitude
    BKG_lambda : float
      background term central wavelength
    BKG_b : float
      background term b coefficient
    BKG_n : float
      background term n coefficient [FIXED at n = 2]

    FUV_amp : float
      far-ultraviolet term amplitude
    FUV_lambda : float
      far-ultraviolet term central wavelength
    FUV_b : float
      far-ultraviolet term b coefficent
    FUV_n : float
      far-ultraviolet term n coefficient

    NUV_amp : float
      near-ultraviolet (2175 A) term amplitude
    NUV_lambda : float
      near-ultraviolet (2175 A) term central wavelength
    NUV_b : float
      near-ultraviolet (2175 A) term b coefficent
    NUV_n : float
      near-ultraviolet (2175 A) term n coefficient [FIXED at n = 2]

    SIL1_amp : float
      1st silicate feature (~10 micron) term amplitude
    SIL1_lambda : float
      1st silicate feature (~10 micron) term central wavelength
    SIL1_b : float
      1st silicate feature (~10 micron) term b coefficent
    SIL1_n : float
      1st silicate feature (~10 micron) term n coefficient [FIXED at n = 2]

    SIL2_amp : float
      2nd silicate feature (~18 micron) term amplitude
    SIL2_lambda : float
      2nd silicate feature (~18 micron) term central wavelength
    SIL2_b : float
      2nd silicate feature (~18 micron) term b coefficient
    SIL2_n : float
      2nd silicate feature (~18 micron) term n coefficient [FIXED at n = 2]

    FIR_amp : float
      far-infrared term amplitude
    FIR_lambda : float
      far-infrared term central wavelength
    FIR_b : float
      far-infrared term b coefficent
    FIR_n : float
      far-infrared term n coefficient [FIXED at n = 2]

    Notes
    -----
    From Pei (1992, ApJ, 395, 130)

    Applicable from the extreme UV to far-IR

    Example showing a P92 curve with components identified.

    .. plot::
        :include-source:

        import numpy as np
        import matplotlib.pyplot as plt
        import astropy.units as u

        from dust_extinction.shapes import P92

        fig, ax = plt.subplots()

        # generate the curves and plot them
        lam = np.logspace(-3.0, 3.0, num=1000)
        x = (1.0/lam)/u.micron

        ext_model = P92()
        ax.plot(1/x,ext_model(x),label='total')

        ext_model = P92(FUV_amp=0., NUV_amp=0.0,
                        SIL1_amp=0.0, SIL2_amp=0.0, FIR_amp=0.0)
        ax.plot(1./x,ext_model(x),label='BKG only')

        ext_model = P92(NUV_amp=0.0,
                        SIL1_amp=0.0, SIL2_amp=0.0, FIR_amp=0.0)
        ax.plot(1./x,ext_model(x),label='BKG+FUV only')

        ext_model = P92(FUV_amp=0.,
                        SIL1_amp=0.0, SIL2_amp=0.0, FIR_amp=0.0)
        ax.plot(1./x,ext_model(x),label='BKG+NUV only')

        ext_model = P92(FUV_amp=0., NUV_amp=0.0,
                        SIL2_amp=0.0)
        ax.plot(1./x,ext_model(x),label='BKG+FIR+SIL1 only')

        ext_model = P92(FUV_amp=0., NUV_amp=0.0,
                        SIL1_amp=0.0)
        ax.plot(1./x,ext_model(x),label='BKG+FIR+SIL2 only')

        ext_model = P92(FUV_amp=0., NUV_amp=0.0,
                        SIL1_amp=0.0, SIL2_amp=0.0)
        ax.plot(1./x,ext_model(x),label='BKG+FIR only')

        # Milky Way observed extinction as tabulated by Pei (1992)
        MW_x = [0.21, 0.29, 0.45, 0.61, 0.80, 1.11, 1.43, 1.82,
                2.27, 2.50, 2.91, 3.65, 4.00, 4.17, 4.35, 4.57, 4.76,
                5.00, 5.26, 5.56, 5.88, 6.25, 6.71, 7.18, 7.60,
                8.00, 8.50, 9.00, 9.50, 10.00]
        MW_x = np.array(MW_x)
        MW_exvebv = [-3.02, -2.91, -2.76, -2.58, -2.23, -1.60, -0.78, 0.00,
                     1.00, 1.30, 1.80, 3.10, 4.19, 4.90, 5.77, 6.57, 6.23,
                     5.52, 4.90, 4.65, 4.60, 4.73, 4.99, 5.36, 5.91,
                     6.55, 7.45, 8.45, 9.80, 11.30]
        MW_exvebv = np.array(MW_exvebv)
        Rv = 3.08
        MW_axav = MW_exvebv/Rv + 1.0
        ax.plot(1./MW_x, MW_axav, 'o', label='MW Observed')

        ax.set_xscale('log')
        ax.set_yscale('log')

        ax.set_ylim(1e-3,10.)

        ax.set_xlabel(r'$\lambda$ [$\mu$m]')
        ax.set_ylabel(r'$A(x)/A(V)$')

        ax.legend(loc='best')
        plt.show()
    """
    n_inputs = 1
    n_outputs = 1

    # constant for conversion from Ax/Ab to (more standard) Ax/Av
    AbAv = 1.0 / 3.08 + 1.0

    BKG_amp = Parameter(
        description="BKG term: amplitude", default=165.0 * AbAv, min=0.0
    )
    BKG_lambda = Parameter(description="BKG term: center wavelength", default=0.047)
    BKG_b = Parameter(description="BKG term: b coefficient", default=90.0)
    BKG_n = Parameter(description="BKG term: n coefficient", default=2.0, fixed=True)

    FUV_amp = Parameter(description="FUV term: amplitude", default=14.0 * AbAv, min=0.0)
    FUV_lambda = Parameter(
        description="FUV term: center wavelength", default=0.07, bounds=(0.06, 0.08)
    )
    FUV_b = Parameter(description="FUV term: b coefficient", default=4.0)
    FUV_n = Parameter(description="FUV term: n coefficient", default=6.5)

    NUV_amp = Parameter(
        description="NUV term: amplitude", default=0.045 * AbAv, min=0.0
    )
    NUV_lambda = Parameter(
        description="NUV term: center wavelength", default=0.22, bounds=(0.20, 0.24)
    )
    NUV_b = Parameter(description="NUV term: b coefficient", default=-1.95)
    NUV_n = Parameter(description="NUV term: n coefficient", default=2.0, fixed=True)

    SIL1_amp = Parameter(
        description="SIL1 term: amplitude", default=0.002 * AbAv, min=0.0
    )
    SIL1_lambda = Parameter(
        description="SIL1 term: center wavelength", default=9.7, bounds=(7.0, 13.0)
    )
    SIL1_b = Parameter(description="SIL1 term: b coefficient", default=-1.95)
    SIL1_n = Parameter(description="SIL1 term: n coefficient", default=2.0, fixed=True)

    SIL2_amp = Parameter(
        description="SIL2 term: amplitude", default=0.002 * AbAv, min=0.0
    )
    SIL2_lambda = Parameter(
        description="SIL2 term: center wavelength", default=18.0, bounds=(15.0, 21.0)
    )
    SIL2_b = Parameter(description="SIL2 term: b coefficient", default=-1.80)
    SIL2_n = Parameter(description="SIL2 term: n coefficient", default=2.0, fixed=True)

    FIR_amp = Parameter(
        description="FIR term: amplitude", default=0.012 * AbAv, min=0.0
    )
    FIR_lambda = Parameter(
        description="FIR term: center wavelength", default=25.0, bounds=(20.0, 30.0)
    )
    FIR_b = Parameter(description="FIR term: b coefficient", default=0.00)
    FIR_n = Parameter(description="FIR term: n coefficient", default=2.0, fixed=True)

    x_range = x_range_P92

    @staticmethod
    def _p92_single_term(in_lambda, amplitude, cen_wave, b, n):
        r"""
        Function for calculating a single P92 term

        .. math::

           \frac{a}{(\lambda/cen_wave)^n + (cen_wave/\lambda)^n + b}

        when n = 2, this term is equivalent to a Drude profile

        Parameters
        ----------
        in_lambda: vector of floats
           wavelengths in same units as cen_wave

        amplitude: float
           amplitude

        cen_wave: flaot
           central wavelength

        b : float
           b coefficient

        n : float
           n coefficient
        """
        l_norm = in_lambda / cen_wave

        return amplitude / (np.power(l_norm, n) + np.power(l_norm, -1 * n) + b)

    def evaluate(
        self,
        in_x,
        BKG_amp,
        BKG_lambda,
        BKG_b,
        BKG_n,
        FUV_amp,
        FUV_lambda,
        FUV_b,
        FUV_n,
        NUV_amp,
        NUV_lambda,
        NUV_b,
        NUV_n,
        SIL1_amp,
        SIL1_lambda,
        SIL1_b,
        SIL1_n,
        SIL2_amp,
        SIL2_lambda,
        SIL2_b,
        SIL2_n,
        FIR_amp,
        FIR_lambda,
        FIR_b,
        FIR_n,
    ):
        """
        P92 function

        Parameters
        ----------
        in_x: float
           expects either x in units of wavelengths or frequency
           or assumes wavelengths in wavenumbers [1/micron]

           internally wavenumbers are used

        Returns
        -------
        axav: np array (float)
            A(x)/A(V) extinction curve [mag]

        Raises
        ------
        ValueError
           Input x values outside of defined range
        """
        x = _get_x_in_wavenumbers(in_x)

        # check that the wavenumbers are within the defined range
        _test_valid_x_range(x, self.x_range, self.__class__.__name__)

        # calculate the terms
        lam = 1.0 / x
        axav = (
            self._p92_single_term(lam, BKG_amp, BKG_lambda, BKG_b, BKG_n)
            + self._p92_single_term(lam, FUV_amp, FUV_lambda, FUV_b, FUV_n)
            + self._p92_single_term(lam, NUV_amp, NUV_lambda, NUV_b, NUV_n)
            + self._p92_single_term(lam, SIL1_amp, SIL1_lambda, SIL1_b, SIL1_n)
            + self._p92_single_term(lam, SIL2_amp, SIL2_lambda, SIL2_b, SIL2_n)
            + self._p92_single_term(lam, FIR_amp, FIR_lambda, FIR_b, FIR_n)
        )

        # return A(x)/A(V)
        return axav

    # use numerical derivaties (need to add analytic)
    fit_deriv = None


class G21(Fittable1DModel):
    r"""
    Gordon et al. (2021) powerlaw plus two modified Drude profiles
    (for the 10 & 20 micron silicate features)
    for the 1 to 40 micron A(lambda)/A(V) extinction curve.

    Parameters
    ----------
    scale: float
        amplitude of the powerlaw at 1 micron
    alpha: float
        power of powerlaw
    sil1_amp: float
        central amplitude of the 10 micron silicate feature
    sil1_center: float
        center wavelength of the 10 micron silicate feature
    sil1_fwhm: float
        full width at half maximum of the 10 micron silicate feature
    sil1_asym: float
        asymmetry of the 10 micron silicate feature
    sil2_amp: float
        central amplitude of the 20 micron silicate feature
    sil2_center: float
        center wavelength of the 20 micron silicate feature
    sil2_fwhm: float
        full width at half maximum of the 20 micron silicate feature
    sil2_asym: float
        asymmetry of the 20 micron silicate feature

    Notes
    -----
    From Gordon et al. (2021, ApJ, submitted)

    Only applicable at NIR/MIR wavelengths from 1-40 micron

    Example showing a G21 curve with components identified.

    .. plot::
        :include-source:

        import numpy as np
        import matplotlib.pyplot as plt
        import astropy.units as u

        from dust_extinction.shapes import G21

        fig, ax = plt.subplots()

        # generate the curves and plot them
        lam = np.logspace(np.log10(1.01), np.log10(39.9), num=1000)
        x = (1.0/lam)/u.micron

        ext_model = G21()
        ax.plot(1/x,ext_model(x),label='total')

        ext_model = G21(sil1_amp=0.0, sil2_amp=0.0)
        ax.plot(1./x,ext_model(x),label='power-law only')

        ext_model = G21(sil2_amp=0.0)
        ax.plot(1./x,ext_model(x),label='power-law+sil1 only')

        ext_model = G21(sil1_amp=0.0)
        ax.plot(1./x,ext_model(x),label='power-law+sil2 only')

        ax.set_xscale('log')
        ax.set_yscale('log')

        ax.set_xlabel('$\lambda$ [$\mu$m]')
        ax.set_ylabel('$A(x)/A(V)$')

        ax.set_title('G21')

        ax.legend(loc='best')
        plt.show()

    """

    # inputs = ("x",)
    # outputs = ("axav",)

    scale = Parameter(
        description="powerlaw: amplitude", default=0.37, bounds=(0.0, 1.0)
    )
    alpha = Parameter(description="powerlaw: alpha", default=1.5, bounds=(0.5, 5.0))
    sil1_amp = Parameter(
        description="silicate 10um: amplitude", default=0.07, bounds=(0.001, 0.3)
    )
    sil1_center = Parameter(
        description="silicate 10um: center", default=9.87, bounds=(8.0, 12.0)
    )
    sil1_fwhm = Parameter(
        description="silicate 10um: fwhm", default=2.5, bounds=(1.0, 10.0)
    )
    sil1_asym = Parameter(
        description="silicate 10um: asymmetry", default=-0.23, bounds=(-2.0, 2.0)
    )
    sil2_amp = Parameter(
        description="silicate 20um: amplitude", default=0.025, bounds=(0.001, 0.3)
    )
    sil2_center = Parameter(
        description="silicate 20um: center", default=17.0, bounds=(16.0, 24.0)
    )
    sil2_fwhm = Parameter(
        description="silicate 20um: fwhm", default=13.0, bounds=(5.0, 20.0)
    )
    sil2_asym = Parameter(
        description="silicate 20um: asymmetry", default=-0.27, bounds=(-2.0, 2.0)
    )

    x_range = [1.0 / 40.0, 1.0]

    def evaluate(
        self,
        in_x,
        scale,
        alpha,
        sil1_amp,
        sil1_center,
        sil1_fwhm,
        sil1_asym,
        sil2_amp,
        sil2_center,
        sil2_fwhm,
        sil2_asym,
    ):
        """
        G21 function

        Parameters
        ----------
        in_x: float
           expects either x in units of wavelengths or frequency
           or assumes wavelengths in wavenumbers [1/micron]

        Returns
        -------
        axav: np array (float)
            A(x)/A(V) extinction curve [mag]

        Raises
        ------
        ValueError
           Input x values outside of defined range
        """
        x = _get_x_in_wavenumbers(in_x)

        # check that the wavenumbers are within the defined range
        _test_valid_x_range(x, self.x_range, "G21")

        wave = 1 / x

        # powerlaw
        axav = scale * (wave ** (-1.0 * alpha))

        # silicate feature drudes
        axav += _modified_drude(wave, sil1_amp, sil1_center, sil1_fwhm, sil1_asym)
        axav += _modified_drude(wave, sil2_amp, sil2_center, sil2_fwhm, sil2_asym)

        return axav



import numpy as np

from astropy.modeling import Model, Parameter, InputParameterError

__all__ = ["BaseExtModel", "BaseExtRvModel", "BaseExtRvAfAModel"]


class BaseExtModel(Model):
    """
    Base Extinction Model.  Do not use.
    """

    n_inputs = 1
    n_outputs = 1

    def extinguish(self, x, Av=None, Ebv=None):
        """
        Calculate the extinction as a fraction

        Parameters
        ----------
        x: float
           expects either x in units of wavelengths or frequency
           or assumes wavelengths in wavenumbers [1/micron]

           internally wavenumbers are used

        Av: float
           A(V) value of dust column
           Av or Ebv must be set

        Ebv: float
           E(B-V) value of dust column
           Av or Ebv must be set

        Returns
        -------
        frac_ext: np array (float)
           fractional extinction as a function of x
        """
        # get the extinction curve
        axav = self(x)

        # check that av or ebv is set
        if (Av is None) and (Ebv is None):
            raise InputParameterError("neither Av or Ebv passed, one required")

        # if Av is not set and Ebv set, convert to Av
        if Av is None:
            Av = self.Rv * Ebv

        # return fractional extinction
        return np.power(10.0, -0.4 * axav * Av)


class BaseExtRvModel(BaseExtModel):
    """
    Base Extinction R(V)-dependent Model.  Do not use.
    """

    Rv = Parameter(
        description="R(V) = A(V)/E(B-V) = " + "total-to-selective extinction",
        default=3.1,
    )

    @Rv.validator
    def Rv(self, value):
        """
        Check that Rv is in the valid range

        Parameters
        ----------
        value: float
            R(V) value to check

        Raises
        ------
        InputParameterError
           Input Rv values outside of defined range
        """
        if not (self.Rv_range[0] <= value <= self.Rv_range[1]):
            raise InputParameterError(
                "parameter Rv must be between "
                + str(self.Rv_range[0])
                + " and "
                + str(self.Rv_range[1])
            )


class BaseExtRvAfAModel(BaseExtModel):
    """
    Base Extinction R(V)_A, f_A -dependent Model.  Do not use.
    """

    RvA = Parameter(
        description="R_A(V) = A(V)/E(B-V) = "
        + "total-to-selective extinction of component A",
        default=3.1,
    )
    fA = Parameter(description="f_A = mixture coefficent of component A", default=1.0)

    def __init__(self, **kwargs):

        super().__init__(**kwargs)

        # set Rv so that extinguishing by Ebv works
        # equation 11 in Gordon et al. (2016, ApJ, 826, 104)
        self.Rv = 1.0 / (self.fA / self.RvA + (1 - self.fA) / 2.74)

    @RvA.validator
    def RvA(self, value):
        """
        Check that RvA is in the valid range

        Parameters
        ----------
        value: float
            RvA value to check

        Raises
        ------
        InputParameterError
           Input R_A(V) values outside of defined range
        """
        if not (self.RvA_range[0] <= value <= self.RvA_range[1]):
            raise InputParameterError(
                "parameter RvA must be between "
                + str(self.RvA_range[0])
                + " and "
                + str(self.RvA_range[1])
            )

    @fA.validator
    def fA(self, value):
        """
        Check that fA is in the valid range

        Parameters
        ----------
        value: float
            fA value to check

        Raises
        ------
        InputParameterError
           Input fA values outside of defined range
        """
        if not (self.fA_range[0] <= value <= self.fA_range[1]):
            raise InputParameterError(
                "parameter fA must be between "
                + str(self.fA_range[0])
                + " and "
                + str(self.fA_range[1])
            )



import pytest

from dust_extinction.parameter_averages import (
    CCM89,
    O94,
    F99,
    F04,
    VCG04,
    GCC09,
    M14,
    G16,
    F19,
    D22,
    G23,
)
from dust_extinction.shapes import FM90, FM90_B3, P92, G21
from dust_extinction.averages import (
    RL85_MWGC,
    RRP89_MWGC,
    B92_MWAvg,
    G03_SMCBar,
    G03_LMCAvg,
    G03_LMC2,
    I05_MWAvg,
    CT06_MWGC,
    CT06_MWLoc,
    GCC09_MWAvg,
    F11_MWGC,
    G21_MWAvg,
    D22_MWAvg,
    G24_SMCAvg,
    G24_SMCBumps,
)
from dust_extinction.grain_models import DBP90, WD01, D03, ZDA04, C11, J13, HD23

param_ave_models_Rv = [CCM89, O94, F99, F04, VCG04, GCC09, M14, F19, D22, G23]
param_ave_models_Rv_fA = [G16]
param_ave_models = param_ave_models_Rv + param_ave_models_Rv_fA
shape_models = [FM90, FM90_B3, P92, G21]
ave_models = [
    RL85_MWGC,
    RRP89_MWGC,
    B92_MWAvg,
    G03_SMCBar,
    G03_LMCAvg,
    G03_LMC2,
    I05_MWAvg,
    CT06_MWGC,
    CT06_MWLoc,
    GCC09_MWAvg,
    F11_MWGC,
    G21_MWAvg,
    D22_MWAvg,
    G24_SMCAvg,
    G24_SMCBumps,
]
grain_models = [DBP90, WD01, D03, ZDA04, C11, J13, HD23]

all_models = (
    param_ave_models_Rv
    + param_ave_models_Rv_fA
    + shape_models
    + ave_models
    + grain_models
)


def _invalid_x_range(x, tmodel, modname):
    with pytest.raises(ValueError) as exc:
        tmodel(x)
    assert (
        exc.value.args[0]
        == "Input x outside of range defined for "
        + modname
        + " ["
        + str(tmodel.x_range[0])
        + " <= x <= "
        + str(tmodel.x_range[1])
        + ", x has units 1/micron]"
    )



import warnings

import numpy as np
from scipy.special import comb
import astropy.units as u

__all__ = ["_get_x_in_wavenumbers", "_test_valid_x_range", "_smoothstep"]


def _get_x_in_wavenumbers(in_x):
    """
    Convert input x to wavenumber given x has units.
    Otherwise, assume x is in waveneumbers and issue a warning to this effect.

    Parameters
    ----------
    in_x : astropy.quantity or simple floats
        x values

    Returns
    -------
    x : floats
        input x values in wavenumbers w/o units
    """
    # handles the case where x is a scaler
    in_x = np.atleast_1d(in_x)

    # check if in_x is an astropy quantity, if not issue a warning
    if not isinstance(in_x, u.Quantity):
        warnings.warn(
            "x has no units, assuming x units are inverse microns", UserWarning
        )

    # convert to wavenumbers (1/micron) if x input in units
    # otherwise, assume x in appropriate wavenumber units
    with u.add_enabled_equivalencies(u.spectral()):
        x_quant = u.Quantity(in_x, 1.0 / u.micron, dtype=np.float64)

    # strip the quantity to avoid needing to add units to all the
    #    polynomical coefficients
    return x_quant.value


def _test_valid_x_range(x, x_range, outname):
    """
    Test if any of the x values are outside of the valid range

    Parameters
    ----------
    x : float array
       wavenumbers in inverse microns

    x_range: 2 floats
       allowed min/max of x

    outname: str
       name of curve for error message
    """
    deltacheck = 1e-6  # delta to allow for small numerical issues
    if np.logical_or(
        np.any(x <= (x_range[0] - deltacheck)), np.any(x >= (x_range[1] + deltacheck))
    ):
        raise ValueError(
            "Input x outside of range defined for "
            + outname
            + " ["
            + str(x_range[0])
            + " <= x <= "
            + str(x_range[1])
            + ", x has units 1/micron]"
        )


def _smoothstep(x, x_min=0, x_max=1, N=1):
    x = np.clip((x - x_min) / (x_max - x_min), 0, 1)

    result = 0
    for n in range(0, N + 1):
        result += comb(N + n, n) * comb(2 * N + 1, N - n) * (-x) ** n

    result *= x ** (N + 1)

    return result



import numpy as np
import pytest

import astropy.units as u
from astropy.modeling.fitting import LevMarLSQFitter

from ..averages import G03_LMCAvg
from ..shapes import FM90


def get_elvebv_cor_vals():
    # testing wavenumbers
    x = np.array([3.5, 4.0, 4.5, 5.0, 6.0, 7.0, 9.0, 10.0])

    # add units
    x = x / u.micron

    # correct values
    cor_vals = np.array(
        [
            2.9829317,
            4.1215415,
            6.4135842,
            5.6574243,
            4.7573250,
            5.4905843,
            9.2853567,
            12.462238,
        ]
    )

    return (x, cor_vals)


def test_extinction_FM90_values():
    # get the correct values
    x, cor_vals = get_elvebv_cor_vals()

    # initialize extinction model
    tmodel = FM90()

    # test
    np.testing.assert_allclose(tmodel(x), cor_vals)


x_vals, axav_vals = get_elvebv_cor_vals()
test_vals = zip(x_vals, axav_vals)


@pytest.mark.parametrize("xtest_vals", test_vals)
def test_extinction_FM90_single_values(xtest_vals):
    x, cor_val = xtest_vals

    # initialize extinction model
    tmodel = FM90()

    # test
    np.testing.assert_allclose(tmodel(x), cor_val)
    np.testing.assert_allclose(
        tmodel.evaluate(
            x,
            FM90.C1.default,
            FM90.C2.default,
            FM90.C3.default,
            FM90.C4.default,
            FM90.xo.default,
            FM90.gamma.default,
        ),
        cor_val,
    )


def test_FM90_fitting():

    # get an observed extinction curve to fit
    g03_model = G03_LMCAvg()

    x = g03_model.obsdata_x
    # convert to E(x-V)/E(B0V)
    y = (g03_model.obsdata_axav - 1.0) * g03_model.Rv
    # only fit the UV portion (FM90 only valid in UV)
    (gindxs,) = np.where(x > 3.125)

    fm90_init = FM90()
    fit = LevMarLSQFitter()
    g03_fit = fit(fm90_init, x[gindxs], y[gindxs])
    fit_vals = [
        g03_fit.C1.value,
        g03_fit.C2.value,
        g03_fit.C3.value,
        g03_fit.C4.value,
        g03_fit.xo.value,
        g03_fit.gamma.value,
    ]

    good_vals = np.array(
        [
            -0.958016797002,
            1.0109751831,
            2.96430606652,
            0.313137860902,
            4.59996300532,
            0.99000982258,
        ]
    )

    np.testing.assert_allclose(good_vals, fit_vals)



import pkg_resources

import numpy as np
from scipy.interpolate import interp1d
from astropy.table import Table
from astropy.modeling.models import PowerLaw1D

from .helpers import _get_x_in_wavenumbers, _test_valid_x_range
from .baseclasses import BaseExtModel
from .shapes import P92, G21, _curve_F99_method

__all__ = [
    "RL85_MWGC",
    "RRP89_MWGC",
    "B92_MWAvg",
    "G03_SMCBar",
    "G03_LMCAvg",
    "G03_LMC2",
    "I05_MWAvg",
    "CT06_MWGC",
    "CT06_MWLoc",
    "GCC09_MWAvg",
    "F11_MWGC",
    "G21_MWAvg",
    "D22_MWAvg",
    "G24_SMCAvg",
    "G24_SMCBumps",
]


class RL85_MWGC(BaseExtModel):
    r"""
    Reike & Lebofsky (1985) MW Average Extinction Curve

    Parameters
    ----------
    None

    Raises
    ------
    None

    Notes
    -----
    From Rieke & Lebofsky (1985, ApJ,288, 618)

    Example showing the average curve

    .. plot::
        :include-source:

        import numpy as np
        import matplotlib.pyplot as plt
        import astropy.units as u

        from dust_extinction.averages import RL85_MWGC

        fig, ax = plt.subplots()

        # define the extinction model
        ext_model = RL85_MWGC()

        # generate the curves and plot them
        x = np.arange(1.0/ext_model.x_range[1], 1.0/ext_model.x_range[0], 0.1) * u.micron

        ax.plot(x,ext_model(x),label='RL85_MWGC')
        ax.plot(1.0/ext_model.obsdata_x, ext_model.obsdata_axav, 'ko',
                label='obsdata')

        ax.set_xlabel(r'$\lambda$ [$\mu m$]')
        ax.set_ylabel(r'$A(x)/A(V)$')

        ax.legend(loc='best')
        plt.show()
    """

    x_range = [1.0 / 13.0, 1.0 / 1.25]

    Rv = 3.09

    # fmt: off
    obsdata_x = 1.0 / np.array(
        [13.0, 12.5, 12.0, 11.5, 11.0, 10.5, 10.0,
         9.5, 9.0, 8.5, 8.0, 4.8, 3.5, 2.22, 1.65, 1.25]
    )
    obsdata_axav = np.array(
        [0.027, 0.030, 0.037, 0.047, 0.060, 0.074, 0.083,
         0.087, 0.074, 0.043, 0.020, 0.023, 0.058, 0.112, 0.175, 0.282]
    )
    # fmt: on

    # accuracy of the observed data based on published table
    obsdata_tolerance = 1e-6

    def evaluate(self, in_x):
        r"""
        RL85 MWGC function

        Parameters
        ----------
        in_x: float
           expects either x in units of wavelengths or frequency
           or assumes wavelengths in wavenumbers [1/micron]

           internally wavenumbers are used

        Returns
        -------
        axav: np array (float)
            A(x)/A(V) extinction curve [mag]

        Raises
        ------
        ValueError
           Input x values outside of defined range
        """
        x = _get_x_in_wavenumbers(in_x)

        # check that the wavenumbers are within the defined range
        _test_valid_x_range(x, self.x_range, self.__class__.__name__)

        # define the function using simple linear interpolation
        # avoids negative values of alav that happens with cubic splines
        f = interp1d(self.obsdata_x, self.obsdata_axav)

        return f(x)


class RRP89_MWGC(BaseExtModel):
    r"""
    Reike, Rieke, & Paul (1989) MW Average Extinction Curve

    Parameters
    ----------
    None

    Raises
    ------
    None

    Notes
    -----
    From Rieke, Rieke, & Paul (1989, ApJ, 336, 752)

    Example showing the average curve

    .. plot::
        :include-source:

        import numpy as np
        import matplotlib.pyplot as plt
        import astropy.units as u

        from dust_extinction.averages import RRP89_MWGC

        fig, ax = plt.subplots()

        # define the extinction model
        ext_model = RRP89_MWGC()

        # generate the curves and plot them
        x = np.arange(1.0/ext_model.x_range[1], 1.0/ext_model.x_range[0], 0.1) * u.micron

        ax.plot(x,ext_model(x),label='RRP89_MWGC')
        ax.plot(1.0/ext_model.obsdata_x, ext_model.obsdata_axav, 'ko',
                label='obsdata')

        ax.set_xlabel(r'$\lambda$ [$\mu m$]')
        ax.set_ylabel(r'$A(x)/A(V)$')

        ax.legend(loc='best')
        plt.show()
    """

    x_range = [1.0 / 13.0, 1.0 / 0.90]

    Rv = 3.09

    # fmt: off
    obsdata_x = 1.0 / np.array(
        [0.90, 1.25, 1.6, 2.2, 3.5, 4.8, 8.0,
         9.5, 10.6, 11.0, 13.0]
    )
    obsdata_elvebv = np.array(
        [-1.60, -2.22, -2.55, -2.744, -2.88, -2.99, -3.01,
         -2.73, -2.87, -2.84, -2.98]
    )
    # fmt: on
    obsdata_axav = obsdata_elvebv / Rv + 1.0

    # accuracy of the observed data based on published table
    obsdata_tolerance = 1e-6

    def evaluate(self, in_x):
        r"""
        RRP89 MWGC function

        Parameters
        ----------
        in_x: float
           expects either x in units of wavelengths or frequency
           or assumes wavelengths in wavenumbers [1/micron]

           internally wavenumbers are used

        Returns
        -------
        axav: np array (float)
            A(x)/A(V) extinction curve [mag]

        Raises
        ------
        ValueError
           Input x values outside of defined range
        """
        x = _get_x_in_wavenumbers(in_x)

        # check that the wavenumbers are within the defined range
        _test_valid_x_range(x, self.x_range, self.__class__.__name__)

        # define the function using simple linear interpolation
        # avoids negative values of alav that happens with cubic splines
        f = interp1d(self.obsdata_x, self.obsdata_axav)

        return f(x)


class B92_MWAvg(BaseExtModel):
    r"""
    Bastiaansen (1992) Optical Extinction Curve

    Parameters
    ----------
    None

    Raises
    ------
    None

    Notes
    -----
    From Bastiaansen (1992, A&AS, 93, 449-462)

    Example showing the average curve

    .. plot::
        :include-source:

        import numpy as np
        import matplotlib.pyplot as plt
        import astropy.units as u

        from dust_extinction.averages import B92_MWAvg

        fig, ax = plt.subplots()

        # define the extinction model
        ext_model = B92_MWAvg()

        # generate the curves and plot them
        x = np.arange(1.0/ext_model.x_range[1], 1.0/ext_model.x_range[0], 0.1) * u.micron

        ax.plot(x,ext_model(x),label='B1992')
        ax.plot(1.0/ext_model.obsdata_x, ext_model.obsdata_axav, 'ko',
                label='obsdata')

        ax.set_xlabel(r'$\lambda$ [$\mu m$]')
        ax.set_ylabel(r'$A(x)/A(V)$')

        ax.legend(loc='best')
        plt.show()
    """

    x_range = [1.0 / 0.7873, 1.0 / 0.3402]

    Rv = 3.1  # assumed!

    # fmt: off
    obsdata_x = 1.0 / np.array(
        [0.7873, 0.7505, 0.7102, 0.6681, 0.64,
         0.6107, 0.5821, 0.5601, 0.5407, 0.5205,
         0.4999, 0.4708, 0.4496, 0.4395, 0.4192,
         0.4038, 0.3785, 0.36, 0.3493, 0.3402]
    )
    obsdata_axav = np.array(
        [0.849, 0.891, 0.941, 0.998, 1.045, 1.088,
         1.139, 1.176, 1.226, 1.279, 1.34 , 1.418,
         1.473, 1.507, 1.556, 1.595, 1.659, 1.718,
         1.761, 1.795]
    )
    # fmt: on

    # accuracy of the observed data based on published table
    obsdata_tolerance = 6e-3

    def evaluate(self, in_x):
        """
        B1992 function

        Parameters
        ----------
        in_x: float
           expects either x in units of wavelengths or frequency
           or assumes wavelengths in wavenumbers [1/micron]

           internally wavenumbers are used

        Returns
        -------
        axav: np array (float)
            A(x)/A(V) extinction curve [mag]

        Raises
        ------
        ValueError
           Input x values outside of defined range
        """

        x = _get_x_in_wavenumbers(in_x)

        # check that the wavenumbers are within the defined range
        _test_valid_x_range(x, self.x_range, self.__class__.name)

        # define the function allowing for spline interpolation
        f = interp1d(self.obsdata_x, self.obsdata_axav)

        return f(x)


class G03_SMCBar(BaseExtModel):
    r"""
    Gordon et al (2003) SMCBar Average Extinction Curve

    Parameters
    ----------
    None

    Raises
    ------
    None

    Notes
    -----
    From Gordon et al. (2003, ApJ, 594, 279)

    The observed A(lambda)/A(V) values at 2.198 and 1.25 microns were
    changed to provide smooth interpolation as noted in
    Gordon et al. (2016, ApJ, 826, 104)

    Example showing the average curve

    .. plot::
        :include-source:

        import numpy as np
        import matplotlib.pyplot as plt
        import astropy.units as u

        from dust_extinction.averages import G03_SMCBar

        fig, ax = plt.subplots()

        # define the extinction model
        ext_model = G03_SMCBar()

        # generate the curves and plot them
        x = np.arange(ext_model.x_range[0], ext_model.x_range[1],0.1)/u.micron

        ax.plot(x,ext_model(x),label='G03 SMCBar')
        ax.plot(ext_model.obsdata_x, ext_model.obsdata_axav, 'ko',
                label='obsdata')

        ax.set_xlabel(r'$x$ [$\mu m^{-1}$]')
        ax.set_ylabel(r'$A(x)/A(V)$')

        ax.legend(loc='best')
        plt.show()
    """

    x_range = [0.3, 10.0]

    Rv = 2.74

    # fmt: off
    obsdata_x = np.array(
        [0.455, 0.606, 0.800, 1.235, 1.538, 1.818, 2.273, 2.703,
         3.375, 3.625, 3.875, 4.125, 4.375, 4.625, 4.875, 5.125,
         5.375, 5.625, 5.875, 6.125, 6.375, 6.625, 6.875, 7.125,
         7.375, 7.625, 7.875, 8.125, 8.375, 8.625]
    )
    obsdata_axav = np.array(
        [0.110, 0.169, 0.250, 0.567, 0.801, 1.000, 1.374, 1.672,
         2.000, 2.220, 2.428, 2.661, 2.947, 3.161, 3.293, 3.489,
         3.637, 3.866, 4.013, 4.243, 4.472, 4.776, 5.000, 5.272,
         5.575, 5.795, 6.074, 6.297, 6.436, 6.992]
    )
    # fmt: on

    # accuracy of the observed data based on published table
    obsdata_tolerance = 6e-2

    def evaluate(self, in_x):
        """
        G03 SMCBar function

        Parameters
        ----------
        in_x: float
           expects either x in units of wavelengths or frequency
           or assumes wavelengths in wavenumbers [1/micron]

           internally wavenumbers are used

        Returns
        -------
        axav: np array (float)
            A(x)/A(V) extinction curve [mag]

        Raises
        ------
        ValueError
           Input x values outside of defined range
        """
        C1 = -4.959
        C2 = 2.264
        C3 = 0.389
        C4 = 0.461
        xo = 4.6
        gamma = 1.0

        optnir_axav_x = 1.0 / np.array(
            [2.198, 1.65, 1.25, 0.81, 0.65, 0.55, 0.44, 0.37]
        )
        # values at 2.198 and 1.25 changed to provide smooth interpolation
        # as noted in Gordon et al. (2016, ApJ, 826, 104)
        optnir_axav_y = [0.11, 0.169, 0.25, 0.567, 0.801, 1.00, 1.374, 1.672]

        # return A(x)/A(V)
        return _curve_F99_method(
            in_x,
            self.Rv,
            C1,
            C2,
            C3,
            C4,
            xo,
            gamma,
            optnir_axav_x,
            optnir_axav_y,
            self.x_range,
            self.__class__.__name__,
        )


class G03_LMCAvg(BaseExtModel):
    r"""
    Gordon et al (2003) LMCAvg Average Extinction Curve

    Parameters
    ----------
    None

    Raises
    ------
    None

    Notes
    -----
    From Gordon et al. (2003, ApJ, 594, 279)

    Example showing the average curve

    .. plot::
        :include-source:

        import numpy as np
        import matplotlib.pyplot as plt
        import astropy.units as u

        from dust_extinction.averages import G03_LMCAvg

        fig, ax = plt.subplots()

        # define the extinction model
        ext_model = G03_LMCAvg()

        # generate the curves and plot them
        x = np.arange(ext_model.x_range[0], ext_model.x_range[1],0.1)/u.micron

        ax.plot(x,ext_model(x),label='G03 LMCAvg')
        ax.plot(ext_model.obsdata_x, ext_model.obsdata_axav, 'ko',
                label='obsdata')

        ax.set_xlabel(r'$x$ [$\mu m^{-1}$]')
        ax.set_ylabel(r'$A(x)/A(V)$')

        ax.legend(loc='best')
        plt.show()
    """

    x_range = [0.3, 10.0]

    Rv = 3.41

    # fmt: off
    obsdata_x = np.array(
        [0.455, 0.606, 0.800, 1.818, 2.273, 2.703, 3.375, 3.625,
         3.875, 4.125, 4.375, 4.625, 4.875, 5.125, 5.375, 5.625,
         5.875, 6.125, 6.375, 6.625, 6.875, 7.125, 7.375, 7.625,
         7.875, 8.125]
    )
    obsdata_axav = np.array(
        [0.100, 0.186, 0.257, 1.000, 1.293, 1.518, 1.786, 1.969,
         2.149, 2.391, 2.771, 2.967, 2.846, 2.646, 2.565, 2.566,
         2.598, 2.607, 2.668, 2.787, 2.874, 2.983, 3.118, 3.231,
         3.374, 3.366]
    )
    # fmt: on

    # accuracy of the observed data based on published table
    obsdata_tolerance = 6e-2

    def evaluate(self, in_x):
        """
        G03 LMCAvg function

        Parameters
        ----------
        in_x: float
           expects either x in units of wavelengths or frequency
           or assumes wavelengths in wavenumbers [1/micron]

           internally wavenumbers are used

        Returns
        -------
        axav: np array (float)
            A(x)/A(V) extinction curve [mag]

        Raises
        ------
        ValueError
           Input x values outside of defined range
        """
        C1 = -0.890
        C2 = 0.998
        C3 = 2.719
        C4 = 0.400
        xo = 4.579
        gamma = 0.934

        optnir_axav_x = 1.0 / np.array([2.198, 1.65, 1.25, 0.55, 0.44, 0.37])
        # value at 2.198 changed to provide smooth interpolation
        # as noted in Gordon et al. (2016, ApJ, 826, 104) for SMCBar
        optnir_axav_y = [0.10, 0.186, 0.257, 1.000, 1.293, 1.518]

        # return A(x)/A(V)
        return _curve_F99_method(
            in_x,
            self.Rv,
            C1,
            C2,
            C3,
            C4,
            xo,
            gamma,
            optnir_axav_x,
            optnir_axav_y,
            self.x_range,
            self.__class__.__name__,
        )


class G03_LMC2(BaseExtModel):
    r"""
    Gordon et al (2003) LMC2 Average Extinction Curve

    Parameters
    ----------
    None

    Raises
    ------
    None

    Notes
    -----
    From Gordon et al. (2003, ApJ, 594, 279)

    Example showing the average curve

    .. plot::
        :include-source:

        import numpy as np
        import matplotlib.pyplot as plt
        import astropy.units as u

        from dust_extinction.averages import G03_LMC2

        fig, ax = plt.subplots()

        # generate the curves and plot them
        x = np.arange(0.3,10.0,0.1)/u.micron

        # define the extinction model
        ext_model = G03_LMC2()

        # generate the curves and plot them
        x = np.arange(ext_model.x_range[0], ext_model.x_range[1],0.1)/u.micron

        ax.plot(x,ext_model(x),label='G03 LMC2')
        ax.plot(ext_model.obsdata_x, ext_model.obsdata_axav, 'ko',
                label='obsdata')

        ax.set_xlabel(r'$x$ [$\mu m^{-1}$]')
        ax.set_ylabel(r'$A(x)/A(V)$')

        ax.legend(loc='best')
        plt.show()
    """

    x_range = [0.3, 10.0]

    Rv = 2.76

    # fmt: off
    obsdata_x = np.array(
        [0.455, 0.606, 0.800, 1.818, 2.273, 2.703, 3.375, 3.625,
         3.875, 4.125, 4.375, 4.625, 4.875, 5.125, 5.375, 5.625,
         5.875, 6.125, 6.375, 6.625, 6.875, 7.125, 7.375, 7.625,
         7.875, 8.125]
    )
    obsdata_axav = np.array(
        [0.101, 0.150, 0.299, 1.000, 1.349, 1.665, 1.899, 2.067,
         2.249, 2.447, 2.777, 2.922, 2.921, 2.812, 2.805, 2.863,
         2.932, 3.060, 3.110, 3.299, 3.408, 3.515, 3.670, 3.862,
         3.937, 4.055]
    )
    # fmt: on

    # accuracy of the observed data based on published table
    obsdata_tolerance = 6e-2

    def evaluate(self, in_x):
        """
        G03 LMC2 function

        Parameters
        ----------
        in_x: float
           expects either x in units of wavelengths or frequency
           or assumes wavelengths in wavenumbers [1/micron]

           internally wavenumbers are used

        Returns
        -------
        axav: np array (float)
            A(x)/A(V) extinction curve [mag]

        Raises
        ------
        ValueError
           Input x values outside of defined range
        """
        C1 = -1.475
        C2 = 1.132
        C3 = 1.463
        C4 = 0.294
        xo = 4.558
        gamma = 0.945

        optnir_axav_x = 1.0 / np.array([2.198, 1.65, 1.25, 0.55, 0.44, 0.37])
        # value at 1.65 changed to provide smooth interpolation
        # as noted in Gordon et al. (2016, ApJ, 826, 104) for SMCBar
        optnir_axav_y = [0.101, 0.15, 0.299, 1.000, 1.349, 1.665]

        # return A(x)/A(V)
        return _curve_F99_method(
            in_x,
            self.Rv,
            C1,
            C2,
            C3,
            C4,
            xo,
            gamma,
            optnir_axav_x,
            optnir_axav_y,
            self.x_range,
            self.__class__.__name__,
        )


class I05_MWAvg(BaseExtModel):
    r"""
    Indebetouw et al (2005) MW Average Extinction Curve

    Parameters
    ----------
    None

    Raises
    ------
    None

    Notes
    -----
    From Indebetouw et al. (2005, ApJ, 619, 931)

    Example showing the average curve

    .. plot::
        :include-source:

        import numpy as np
        import matplotlib.pyplot as plt
        import astropy.units as u

        from dust_extinction.averages import I05_MWAvg

        fig, ax = plt.subplots()

        # define the extinction model
        ext_model = I05_MWAvg()

        # generate the curves and plot them
        x = np.arange(1.0/ext_model.x_range[1], 1.0/ext_model.x_range[0], 0.1) * u.micron

        ax.plot(x,ext_model(x),label='I05_MWAvg')
        ax.plot(1.0/ext_model.obsdata_x, ext_model.obsdata_axav, 'ko',
                label='obsdata')

        ax.set_xlabel(r'$\lambda$ [$\mu m$]')
        ax.set_ylabel(r'$A(x)/A(V)$')

        ax.legend(loc='best')
        plt.show()
    """

    x_range = [1.0 / 7.76, 1.0 / 1.24]

    Rv = 3.1  # assumed!

    # fmt: off
    obsdata_x = 1.0 / np.array(
        [1.24, 1.664, 2.164, 3.545, 4.442, 5.675, 7.760]
    )
    obsdata_axav = np.array(
        [2.50, 1.55, 1.00, 0.56, 0.43, 0.43, 0.43]
    ) * 0.112  # ak/av = 0.112 (F19, Rv = 3.1)

    obsdata_axav_unc = np.array(
        [0.15, 0.08, 0.0, 0.06, 0.08, 0.10, 0.10]
    ) * 0.112  # ak/av = 0.112 (F19, Rv = 3.1)
    # fmt: on

    # accuracy of the observed data based on published table
    obsdata_tolerance = 1e-6

    def evaluate(self, in_x):
        """
        I05 MWAvg function

        Parameters
        ----------
        in_x: float
           expects either x in units of wavelengths or frequency
           or assumes wavelengths in wavenumbers [1/micron]

           internally wavenumbers are used

        Returns
        -------
        axav: np array (float)
            A(x)/A(V) extinction curve [mag]

        Raises
        ------
        ValueError
           Input x values outside of defined range
        """
        x = _get_x_in_wavenumbers(in_x)

        # check that the wavenumbers are within the defined range
        _test_valid_x_range(x, self.x_range, self.__class__.__name__)

        # define the function allowing for spline interpolation
        f = interp1d(self.obsdata_x, self.obsdata_axav)

        return f(x)


class CT06_MWGC(BaseExtModel):
    r"""
    Chiar & Tielens (2006) MW Galactic Center Curve

    Parameters
    ----------
    None

    Raises
    ------
    None

    Notes
    -----
    From Chiar & Tielens (2006, ApJ, 637 774)

    Example showing the average curve

    .. plot::
        :include-source:

        import numpy as np
        import matplotlib.pyplot as plt
        import astropy.units as u

        from dust_extinction.averages import CT06_MWGC

        fig, ax = plt.subplots()

        # define the extinction model
        ext_model = CT06_MWGC()

        # generate the curves and plot them
        x = np.arange(1.0/ext_model.x_range[1], 1.0/ext_model.x_range[0], 0.1) * u.micron

        ax.plot(x,ext_model(x),label='CT06_MWGC')
        ax.plot(1.0/ext_model.obsdata_x, ext_model.obsdata_axav, 'ko',
                label='obsdata')

        ax.set_xlabel(r'$\lambda$ [$\mu m$]')
        ax.set_ylabel(r'$A(x)/A(V)$')

        ax.legend(loc='best')
        plt.show()
    """

    x_range = [1.0 / 27.0, 1.0 / 1.24]

    Rv = 3.1  # assumed!

    def __init__(self, **kwargs):

        # get the tabulated information
        data_path = pkg_resources.resource_filename("dust_extinction", "data/")

        a = Table.read(
            data_path + "CT06_pixiedust.dat", format="ascii.commented_header"
        )

        self.obsdata_x = 1.0 / a["wave"].data
        # ext is A(lambda)/A(K)
        # A(K)/A(V) = 0.112 (F19, R(V) = 3.1)
        self.obsdata_axav = 0.112 * a["galcen"].data

        # accuracy of the observed data based on published table
        self.obsdata_tolerance = 1e-6

        super().__init__(**kwargs)

    def evaluate(self, in_x):
        """
        CT06 MWGC function

        Parameters
        ----------
        in_x: float
           expects either x in units of wavelengths or frequency
           or assumes wavelengths in wavenumbers [1/micron]

           internally wavenumbers are used

        Returns
        -------
        axav: np array (float)
            A(x)/A(V) extinction curve [mag]

        Raises
        ------
        ValueError
           Input x values outside of defined range
        """
        x = _get_x_in_wavenumbers(in_x)

        # check that the wavenumbers are within the defined range
        _test_valid_x_range(x, self.x_range, self.__class__.__name__)

        # define the function allowing for spline interpolation
        f = interp1d(self.obsdata_x, self.obsdata_axav)

        return f(x)


class CT06_MWLoc(BaseExtModel):
    r"""
    Chiar & Tielens (2006) MW Local ISM Curve

    Parameters
    ----------
    None

    Raises
    ------
    None

    Notes
    -----
    From Chiar & Tielens (2006, ApJ, 637 774)

    Example showing the average curve

    .. plot::
        :include-source:

        import numpy as np
        import matplotlib.pyplot as plt
        import astropy.units as u

        from dust_extinction.averages import CT06_MWLoc

        fig, ax = plt.subplots()

        # define the extinction model
        ext_model = CT06_MWLoc()

        # generate the curves and plot them
        x = np.arange(1.0/ext_model.x_range[1], 1.0/ext_model.x_range[0], 0.1) * u.micron

        ax.plot(x,ext_model(x),label='CT06_MWLoc')
        ax.plot(1.0/ext_model.obsdata_x, ext_model.obsdata_axav, 'ko',
                label='obsdata')

        ax.set_xlabel(r'$\lambda$ [$\mu m$]')
        ax.set_ylabel(r'$A(x)/A(V)$')

        ax.legend(loc='best')
        plt.show()
    """

    x_range = [1.0 / 27.0, 1.0 / 1.24]

    Rv = 3.1  # assumed!

    def __init__(self, **kwargs):

        # get the tabulated information
        data_path = pkg_resources.resource_filename("dust_extinction", "data/")

        a = Table.read(
            data_path + "CT06_pixiedust.dat", format="ascii.commented_header"
        )

        self.obsdata_x = 1.0 / a["wave"].data
        # ext is A(lambda)/A(K)
        # A(K)/A(V) = 0.112 (F19, R(V) = 3.1)
        self.obsdata_axav = 0.112 * a["local"].data

        # accuracy of the observed data based on published table
        self.obsdata_tolerance = 1e-6

        super().__init__(**kwargs)

    def evaluate(self, in_x):
        """
        CG06 MWLoc function

        Parameters
        ----------
        in_x: float
           expects either x in units of wavelengths or frequency
           or assumes wavelengths in wavenumbers [1/micron]

           internally wavenumbers are used

        Returns
        -------
        axav: np array (float)
            A(x)/A(V) extinction curve [mag]

        Raises
        ------
        ValueError
           Input x values outside of defined range
        """
        x = _get_x_in_wavenumbers(in_x)

        # check that the wavenumbers are within the defined range
        _test_valid_x_range(x, self.x_range, self.__class__.__name__)

        # define the function allowing for spline interpolation
        f = interp1d(self.obsdata_x, self.obsdata_axav)

        return f(x)


class GCC09_MWAvg(BaseExtModel):
    r"""
    Gordon, Cartledge, & Clayton (2009) Milky Way Average Extinction Curve

    Parameters
    ----------
    None

    Raises
    ------
    None

    Notes
    -----
    From Gordon, Cartledge, & Clayton (2009, ApJ, 705, 1320)

    Example showing the average curve

    .. plot::
        :include-source:

        import numpy as np
        import matplotlib.pyplot as plt
        import astropy.units as u

        from dust_extinction.averages import GCC09_MWAvg

        fig, ax = plt.subplots()

        # generate the curves and plot them
        x = np.arange(0.3,1.0/0.0912,0.1)/u.micron

        # define the extinction model
        ext_model = GCC09_MWAvg()

        # generate the curves and plot them
        x = np.arange(ext_model.x_range[0], ext_model.x_range[1],0.1)/u.micron

        ax.plot(x,ext_model(x),label='GCC09_MWAvg')
        ax.errorbar(ext_model.obsdata_x_fuse, ext_model.obsdata_axav_fuse,
                    yerr=ext_model.obsdata_axav_unc_fuse,
                    fmt='ko', label='obsdata (FUSE)')
        ax.errorbar(ext_model.obsdata_x_iue, ext_model.obsdata_axav_iue,
                    yerr=ext_model.obsdata_axav_unc_iue,
                    fmt='bs', label='obsdata (IUE)')
        ax.errorbar(ext_model.obsdata_x_bands, ext_model.obsdata_axav_bands,
                    yerr=ext_model.obsdata_axav_unc_bands,
                    fmt='g^', label='obsdata (Opt/NIR)')

        ax.set_xlabel(r'$x$ [$\mu m^{-1}$]')
        ax.set_ylabel(r'$A(x)/A(V)$')

        ax.legend(loc='best')
        plt.show()
    """

    x_range = [0.3, 1.0 / 0.0912]

    Rv = 3.1

    def __init__(self, **kwargs):

        # get the tabulated information
        data_path = pkg_resources.resource_filename("dust_extinction", "data/")

        # GCC09 sigma clipped average of 75 sightlines
        a = Table.read(data_path + "GCC09_FUSE.dat", format="ascii.commented_header")
        b = Table.read(data_path + "GCC09_IUE.dat", format="ascii.commented_header")
        c = Table.read(data_path + "GCC09_PHOT.dat", format="ascii.commented_header")

        # FUSE range
        self.obsdata_x_fuse = a["x"].data
        self.obsdata_axav_fuse = a["ext"].data
        self.obsdata_axav_unc_fuse = a["unc"].data

        # IUE range
        self.obsdata_x_iue = b["x"].data
        self.obsdata_axav_iue = b["ext"].data
        self.obsdata_axav_unc_iue = b["unc"].data

        # Opt/NIR range
        self.obsdata_x_bands = c["x"].data
        self.obsdata_axav_bands = c["ext"].data
        self.obsdata_axav_unc_bands = c["unc"].data

        # put them together
        self.obsdata_x = np.concatenate(
            (self.obsdata_x_fuse, self.obsdata_x_iue, self.obsdata_x_bands)
        )
        self.obsdata_axav = np.concatenate(
            (self.obsdata_axav_fuse, self.obsdata_axav_iue, self.obsdata_axav_bands)
        )
        self.obsdata_axav_unc = np.concatenate(
            (
                self.obsdata_axav_unc_fuse,
                self.obsdata_axav_unc_iue,
                self.obsdata_axav_unc_bands,
            )
        )

        # accuracy of the observed data based on published table
        self.obsdata_tolerance = 5e-1

        super().__init__(**kwargs)

    def evaluate(self, in_x):
        """
        GCC09_MWAvg function

        Parameters
        ----------
        in_x: float
           expects either x in units of wavelengths or frequency
           or assumes wavelengths in wavenumbers [1/micron]

           internally wavenumbers are used

        Returns
        -------
        axav: np array (float)
            A(x)/A(V) extinction curve [mag]

        Raises
        ------
        ValueError
           Input x values outside of defined range
        """
        x = _get_x_in_wavenumbers(in_x)

        # check that the wavenumbers are within the defined range
        _test_valid_x_range(x, self.x_range, self.__class__.__name__)

        # P92 parameters fit to the data using uncs as weights
        p92_fit = P92(
            BKG_amp=203.805939127,
            BKG_lambda=0.0508199427208,
            BKG_b=88.0591826413,
            BKG_n=2.0,
            FUV_amp=5.33962141873,
            FUV_lambda=0.08,
            FUV_b=-0.777129536415,
            FUV_n=3.88322376926,
            NUV_amp=0.0447023090042,
            NUV_lambda=0.217548391182,
            NUV_b=-1.95723797612,
            NUV_n=2.0,
            SIL1_amp=0.00264935064935,
            SIL1_lambda=9.7,
            SIL1_b=-1.95,
            SIL1_n=2.0,
            SIL2_amp=0.00264935064935,
            SIL2_lambda=18.0,
            SIL2_b=-1.80,
            SIL2_n=2.0,
            FIR_amp=0.01589610389,
            FIR_lambda=25.0,
            FIR_b=0.0,
            FIR_n=2.0,
        )

        # return A(x)/A(V)
        return p92_fit(in_x)


class F11_MWGC(BaseExtModel):
    r"""
    Fritz et al (2011) MW Galactic Center Curve

    Parameters
    ----------
    None

    Raises
    ------
    None

    Notes
    -----
    From Fritz et al. (2011, ApJ, 737, 73)

    Example showing the average curve

    .. plot::
        :include-source:

        import numpy as np
        import matplotlib.pyplot as plt
        import astropy.units as u

        from dust_extinction.averages import F11_MWGC

        fig, ax = plt.subplots()

        # define the extinction model
        ext_model = F11_MWGC()

        # generate the curves and plot them
        x = np.arange(1.0/ext_model.x_range[1], 1.0/ext_model.x_range[0], 0.1) * u.micron

        ax.plot(x,ext_model(x),label='F11_MWGC')
        ax.plot(1.0/ext_model.obsdata_x, ext_model.obsdata_axav, 'ko',
                label='obsdata')

        ax.set_xlabel(r'$\lambda$ [$\mu m$]')
        ax.set_ylabel(r'$A(x)/A(V)$')

        ax.legend(loc='best')
        plt.show()
    """

    x_range = [1.0 / 19.062, 1.0 / 1.282]

    Rv = 3.1  # assumed!

    def __init__(self, **kwargs):

        # get the tabulated information
        data_path = pkg_resources.resource_filename("dust_extinction", "data/")

        a = Table.read(
            data_path + "fritz11_galcenter.dat", format="ascii.commented_header"
        )

        self.obsdata_x = 1.0 / a["wave"].data
        # ext is total extinction to GalCenter
        # A(K) = 2.42
        # A(K)/A(V) = 0.112 (F19, R(V) = 3.1)
        self.obsdata_axav = 0.112 * a["ext"].data / 2.42
        self.obsdata_axav_unc = 0.112 * a["unc"].data / 2.42

        # accuracy of the observed data based on published table
        self.obsdata_tolerance = 1e-6

        super().__init__(**kwargs)

    def evaluate(self, in_x):
        """
        F11 MWGC function

        Parameters
        ----------
        in_x: float
           expects either x in units of wavelengths or frequency
           or assumes wavelengths in wavenumbers [1/micron]

           internally wavenumbers are used

        Returns
        -------
        axav: np array (float)
            A(x)/A(V) extinction curve [mag]

        Raises
        ------
        ValueError
           Input x values outside of defined range
        """
        x = _get_x_in_wavenumbers(in_x)

        # check that the wavenumbers are within the defined range
        _test_valid_x_range(x, self.x_range, self.__class__.__name__)

        # define the function allowing for spline interpolation
        f = interp1d(self.obsdata_x, self.obsdata_axav)

        return f(x)


class G21_MWAvg(BaseExtModel):
    r"""
    Gordon et al. (2021) Milky Way Average Extinction Curve

    Parameters
    ----------
    None

    Raises
    ------
    None

    Notes
    -----
    From Gordon et al. (2021, ApJ, 916, 33)

    Example showing the average curve

    .. plot::
        :include-source:

        import numpy as np
        import matplotlib.pyplot as plt
        import astropy.units as u

        from dust_extinction.averages import G21_MWAvg

        fig, ax = plt.subplots()

        # generate the curves and plot them
        lam = np.logspace(np.log10(1.01), np.log10(31.9), num=1000)
        x = (1.0/lam)/u.micron

        # define the extinction model
        ext_model = G21_MWAvg()

        ax.plot(1.0/x,ext_model(x),label='G21_MWAvg')
        ax.errorbar(1.0/ext_model.obsdata_x_irs, ext_model.obsdata_axav_irs,
                    yerr=ext_model.obsdata_axav_unc_irs,
                    fmt='ko', label='obsdata (IRS)')
        ax.errorbar(1.0/ext_model.obsdata_x_bands, ext_model.obsdata_axav_bands,
                    yerr=ext_model.obsdata_axav_unc_bands,
                    fmt='g^', label='obsdata (photometry)')

        ax.set_xlabel(r'$\lambda$ [$\mu m$]')
        ax.set_ylabel(r'$A(x)/A(V)$')

        ax.set_xscale('log')
        ax.set_yscale('log')

        ax.legend(loc='best')
        plt.show()
    """

    x_range = [1.0 / 32.0, 1.0]

    Rv = 3.17

    def __init__(self, **kwargs):

        # get the tabulated information
        data_path = pkg_resources.resource_filename("dust_extinction", "data/")

        # GCC09 sigma clipped average of 75 sightlines
        a = Table.read(data_path + "G21_IRS.dat", format="ascii.commented_header")
        b = Table.read(data_path + "G21_PHOT.dat", format="ascii.commented_header")

        # IRS range
        self.obsdata_x_irs = 1.0 / a["wave"].data
        self.obsdata_axav_irs = a["ext"].data
        self.obsdata_axav_unc_irs = a["unc"].data

        # Opt/NIR range
        self.obsdata_x_bands = 1.0 / b["wave"].data
        self.obsdata_axav_bands = b["ext"].data
        self.obsdata_axav_unc_bands = b["unc"].data

        # put them together
        self.obsdata_x = np.concatenate((self.obsdata_x_irs, self.obsdata_x_bands))
        self.obsdata_axav = np.concatenate(
            (self.obsdata_axav_irs, self.obsdata_axav_bands)
        )
        self.obsdata_axav_unc = np.concatenate(
            (
                self.obsdata_axav_unc_irs,
                self.obsdata_axav_unc_bands,
            )
        )

        # accuracy of the observed data based on published table
        self.obsdata_tolerance = 5e-1

        super().__init__(**kwargs)

    def evaluate(self, in_x):
        """
        G21_MWAvg function

        Parameters
        ----------
        in_x: float
           expects either x in units of wavelengths or frequency
           or assumes wavelengths in wavenumbers [1/micron]

           internally wavenumbers are used

        Returns
        -------
        axav: np array (float)
            A(x)/A(V) extinction curve [mag]

        Raises
        ------
        ValueError
           Input x values outside of defined range
        """
        x = _get_x_in_wavenumbers(in_x)

        # check that the wavenumbers are within the defined range
        _test_valid_x_range(x, self.x_range, self.__class__.__name__)

        # G21 parameters fit to the data using uncs as weights
        g21_fit = G21(
            scale=0.366,
            alpha=1.480,
            sil1_amp=0.06893,
            sil1_center=9.865,
            sil1_fwhm=2.507,
            sil1_asym=-0.232,
            sil2_amp=0.02684,
            sil2_center=19.973,
            sil2_fwhm=16.989,
            sil2_asym=-0.273,
        )

        # return A(x)/A(V)
        # G21 a full dust_extinction model, hence send in x with units
        return g21_fit(in_x)


class D22_MWAvg(BaseExtModel):
    r"""
    Decleir et al. (2022) Milky Way Average Extinction Curve

    Parameters
    ----------
    None

    Raises
    ------
    None

    Notes
    -----
    From Decleir et al. (2022, ApJ, submitted)

    Example showing the average curve

    .. plot::
        :include-source:

        import numpy as np
        import matplotlib.pyplot as plt
        import astropy.units as u

        from dust_extinction.averages import D22_MWAvg

        fig, ax = plt.subplots()

        # generate the curves and plot them
        lam = np.logspace(np.log10(0.8), np.log10(4.9), num=1000)
        x = (1.0 / lam) / u.micron

        # define the extinction model
        ext_model = D22_MWAvg()

        ax.plot(1.0 / x, ext_model(x), label="D22_MWAvg")
        ax.errorbar(
            1.0 / ext_model.obsdata_x,
            ext_model.obsdata_axav,
            yerr=ext_model.obsdata_axav_unc,
            fmt="ko",
            label="obsdata",
        )

        ax.set_xlabel(r"$\lambda$ [$\mu m$]")
        ax.set_ylabel(r"$A(x)/A(V)$")

        ax.set_xscale("log")
        ax.set_yscale("log")

        ax.legend(loc="best")
        plt.show()

    """

    x_range = [1.0 / 5.0, 1.0 / 0.8]

    Rv = 3.12

    def __init__(self, **kwargs):

        # get the tabulated information
        data_path = pkg_resources.resource_filename("dust_extinction", "data/")

        # D22 sigma clipped average of 13 diffuse sightlines
        a = Table.read(data_path + "D22.dat", format="ascii.commented_header")

        # Spex data
        self.obsdata_x = 1.0 / a["wavelength[micron]"].data
        self.obsdata_axav = a["ave"].data
        self.obsdata_axav_unc = a["ave_unc"].data

        # accuracy of the observed data based on published table
        self.obsdata_tolerance = 0.2  # check

        super().__init__(**kwargs)

    def evaluate(self, in_x):
        """
        D22_MWAvg function

        Parameters
        ----------
        in_x: float
           expects either x in units of wavelengths or frequency
           or assumes wavelengths in wavenumbers [1/micron]

           internally wavenumbers are used

        Returns
        -------
        axav: np array (float)
            A(x)/A(V) extinction curve [mag]

        Raises
        ------
        ValueError
           Input x values outside of defined range
        """
        x = _get_x_in_wavenumbers(in_x)

        # check that the wavenumbers are within the defined range
        _test_valid_x_range(x, self.x_range, self.__class__.__name__)

        # setup the model
        d22_fit = PowerLaw1D(alpha=1.71, amplitude=0.386, x_0=1.0)

        # return A(x)/A(V)
        # Note that model in D22 was done versus wavelength in microns
        return d22_fit(1.0 / x)


class G24_SMCAvg(BaseExtModel):
    r"""
    Gordon et al (2024) SMC Average Extinction Curve

    Parameters
    ----------
    None

    Raises
    ------
    None

    Notes
    -----
    From Gordon et al. (2024, ApJ, in press)

    Example showing the average curve

    .. plot::
        :include-source:

        import numpy as np
        import matplotlib.pyplot as plt
        import astropy.units as u

        from dust_extinction.averages import G24_SMCAvg

        fig, ax = plt.subplots()

        # define the extinction model
        ext_model = G24_SMCAvg()

        # generate the curves and plot them
        x = np.arange(ext_model.x_range[0], ext_model.x_range[1],0.1)/u.micron

        ax.plot(x,ext_model(x),label='G24 SMC Average')
        ax.plot(ext_model.obsdata_x, ext_model.obsdata_axav, 'ko',
                label='obsdata')

        ax.set_xlabel(r'$x$ [$\mu m^{-1}$]')
        ax.set_ylabel(r'$A(x)/A(V)$')

        ax.legend(loc='best')
        plt.show()
    """

    x_range = [0.3, 10.0]

    Rv = 3.02

    def __init__(self, **kwargs):

        # get the tabulated information
        data_path = pkg_resources.resource_filename("dust_extinction", "data/")

        # D22 sigma clipped average of 13 diffuse sightlines
        a = Table.read(data_path + "G24_SMCAvg.dat", format="ascii.commented_header")

        # data
        self.obsdata_x = 1.0 / a["wave"].data
        self.obsdata_axav = a["A(l)/A(V)"].data
        self.obsdata_axav_unc = a["unc"].data

        # accuracy of the observed data based on published table
        self.obsdata_tolerance = (
            0.1  # large value driven by short wavelength uncertainties
        )

        super().__init__(**kwargs)

    def evaluate(self, in_x):
        """
        G24 SMCAvg function

        Parameters
        ----------
        in_x: float
           expects either x in units of wavelengths or frequency
           or assumes wavelengths in wavenumbers [1/micron]

           internally wavenumbers are used

        Returns
        -------
        axav: np array (float)
            A(x)/A(V) extinction curve [mag]

        Raises
        ------
        ValueError
           Input x values outside of defined range
        """
        C1 = -5.07
        C2 = 2.30
        C3 = 0.12
        C4 = 0.07
        xo = 4.59
        gamma = 0.95

        optnir_axav_x = 1.0 / np.array([2.198, 1.65, 1.25, 0.55, 0.44, 0.37])
        optnir_axav_y = [0.062, 0.125, 0.324, 1.021, 1.349, 1.514]

        # return A(x)/A(V)
        return _curve_F99_method(
            in_x,
            self.Rv,
            C1,
            C2,
            C3,
            C4,
            xo,
            gamma,
            optnir_axav_x,
            optnir_axav_y,
            self.x_range,
            self.__class__.__name__,
        )


class G24_SMCBumps(BaseExtModel):
    r"""
    Gordon et al (2024) SMC Bumps Extinction Curve

    Parameters
    ----------
    None

    Raises
    ------
    None

    Notes
    -----
    From Gordon et al. (2024, ApJ, in press)

    Two data points in the FUV from the data file giving the observed average were removed
    as they are *very* deviate from the FM90 parametrization.  This cause the automated tests
    to fail.

    Example showing the average curve

    .. plot::
        :include-source:

        import numpy as np
        import matplotlib.pyplot as plt
        import astropy.units as u

        from dust_extinction.averages import G24_SMCBumps

        fig, ax = plt.subplots()

        # define the extinction model
        ext_model = G24_SMCBumps()

        # generate the curves and plot them
        x = np.arange(ext_model.x_range[0], ext_model.x_range[1],0.1)/u.micron

        ax.plot(x,ext_model(x),label='G24 SMC Bumps')
        ax.plot(ext_model.obsdata_x, ext_model.obsdata_axav, 'ko',
                label='obsdata')

        ax.set_xlabel(r'$x$ [$\mu m^{-1}$]')
        ax.set_ylabel(r'$A(x)/A(V)$')

        ax.legend(loc='best')
        plt.show()
    """

    x_range = [0.3, 10.0]

    Rv = 2.55

    def __init__(self, **kwargs):

        # get the tabulated information
        data_path = pkg_resources.resource_filename("dust_extinction", "data/")

        # D22 sigma clipped average of 13 diffuse sightlines
        a = Table.read(data_path + "G24_SMCBumps.dat", format="ascii.commented_header")

        # data
        self.obsdata_x = 1.0 / a["wave"].data
        self.obsdata_axav = a["A(l)/A(V)"].data
        self.obsdata_axav_unc = a["unc"].data

        # accuracy of the observed data based on published table
        self.obsdata_tolerance = (
            0.1  # large value driven by short wavelength uncertainties
        )

        super().__init__(**kwargs)

    def evaluate(self, in_x):
        """
        G24 SMCBumps function

        Parameters
        ----------
        in_x: float
           expects either x in units of wavelengths or frequency
           or assumes wavelengths in wavenumbers [1/micron]

           internally wavenumbers are used

        Returns
        -------
        axav: np array (float)
            A(x)/A(V) extinction curve [mag]

        Raises
        ------
        ValueError
           Input x values outside of defined range
        """
        C1 = -2.85
        C2 = 1.51
        C3 = 2.64
        C4 = 0.25
        xo = 4.73
        gamma = 1.15

        optnir_axav_x = 1.0 / np.array([2.198, 1.65, 1.25, 0.55, 0.44, 0.37])
        optnir_axav_y = [0.055, 0.162, 0.293, 1.017, 1.400, 1.684]

        # return A(x)/A(V)
        return _curve_F99_method(
            in_x,
            self.Rv,
            C1,
            C2,
            C3,
            C4,
            xo,
            gamma,
            optnir_axav_x,
            optnir_axav_y,
            self.x_range,
            self.__class__.__name__,
        )



import numpy as np
import pytest

import astropy.units as u
from astropy.modeling.fitting import LevMarLSQFitter

from ..shapes import P92


def get_axav_cor_vals():

    # Milky Way observed extinction as tabulated by Pei (1992)
    MW_x = [
        0.21,
        0.29,
        0.45,
        0.61,
        0.80,
        1.11,
        1.43,
        1.82,
        2.27,
        2.50,
        2.91,
        3.65,
        4.00,
        4.17,
        4.35,
        4.57,
        4.76,
        5.00,
        5.26,
        5.56,
        5.88,
        6.25,
        6.71,
        7.18,
        7.60,
        8.00,
        8.50,
        9.00,
        9.50,
        10.00,
    ]
    MW_x = np.array(MW_x)
    MW_exvebv = [
        -3.02,
        -2.91,
        -2.76,
        -2.58,
        -2.23,
        -1.60,
        -0.78,
        0.00,
        1.00,
        1.30,
        1.80,
        3.10,
        4.19,
        4.90,
        5.77,
        6.57,
        6.23,
        5.52,
        4.90,
        4.65,
        4.60,
        4.73,
        4.99,
        5.36,
        5.91,
        6.55,
        7.45,
        8.45,
        9.80,
        11.30,
    ]
    MW_exvebv = np.array(MW_exvebv)
    Rv = 3.08
    MW_axav = MW_exvebv / Rv + 1.0

    # add units
    x = MW_x / u.micron

    # correct values
    cor_vals = MW_axav

    return (x, cor_vals)


def test_extinction_P92_values():
    # get the correct values
    x, cor_vals = get_axav_cor_vals()

    # initialize extinction model
    tmodel = P92()

    # test
    np.testing.assert_allclose(tmodel(x), cor_vals, rtol=0.25, atol=0.01)


x_vals, axav_vals = get_axav_cor_vals()
test_vals = zip(x_vals, axav_vals)


@pytest.mark.parametrize("xtest_vals", test_vals)
def test_extinction_P92_single_values(xtest_vals):
    x, cor_val = xtest_vals

    # initialize extinction model
    tmodel = P92()

    # test
    np.testing.assert_allclose(tmodel(x), cor_val, rtol=0.25, atol=0.01)
    np.testing.assert_allclose(
        tmodel.evaluate(
            x,
            P92.BKG_amp.default,
            P92.BKG_lambda.default,
            P92.BKG_b.default,
            P92.BKG_n.default,
            P92.FUV_amp.default,
            P92.FUV_lambda.default,
            P92.FUV_b.default,
            P92.FUV_n.default,
            P92.NUV_amp.default,
            P92.NUV_lambda.default,
            P92.NUV_b.default,
            P92.NUV_n.default,
            P92.SIL1_amp.default,
            P92.SIL1_lambda.default,
            P92.SIL1_b.default,
            P92.SIL1_n.default,
            P92.SIL2_amp.default,
            P92.SIL2_lambda.default,
            P92.SIL2_b.default,
            P92.SIL2_n.default,
            P92.FIR_amp.default,
            P92.FIR_lambda.default,
            P92.FIR_b.default,
            P92.FIR_n.default,
        ),
        cor_val,
        rtol=0.25,
        atol=0.01,
    )


@pytest.mark.skip(reason="failing due to an issue with the fitting")
def test_P92_fitting():

    # get an observed extinction curve to fit
    x_quant, y = get_axav_cor_vals()

    x = x_quant.value

    # change from defaults to make the best fit harder to find
    p92_init = P92()

    fit = LevMarLSQFitter()
    # accuracy set to avoid warning the fit may have failed
    p92_fit = fit(p92_init, x, y, acc=1e-2)

    fit_vals = p92_fit._parameters

    good_vals = [
        218.957451206,
        0.0481323587043,
        89.8639079339,
        2.0,
        19.8918861271,
        0.0674934514694,
        0.919702726068,
        5.1217891448,
        0.0548568919776,
        0.218664938289,
        -1.9496661308,
        2.0,
        0.0,
        13.0,
        38.279150331,
        2.0,
        0.0,
        15.0,
        -76.7467816812,
        2.0,
        0.0,
        20.0,
        -2508.60124085,
        2.0,
    ]

    np.testing.assert_allclose(good_vals, fit_vals)



import numpy as np
import pytest

import astropy.units as u

from ..parameter_averages import GCC09


def get_axav_cor_vals(Rv):
    # testing wavenumbers
    x = np.array([10.0, 9.0, 8.0, 7.0, 6.0, 5.0, 4.6, 4.0, 3.4])

    # add units
    x = x / u.micron

    # correct values
    # from IDL version
    if Rv == 3.1:
        cor_vals = np.array(
            [
                5.23161,
                4.20810,
                3.45123,
                2.92264,
                2.61283,
                2.85130,
                3.19451,
                2.34301,
                1.89256,
            ]
        )
    elif Rv == 2.0:
        cor_vals = np.array(
            [
                10.5150,
                8.07274,
                6.26711,
                5.00591,
                4.24237,
                4.42844,
                4.99482,
                3.42585,
                2.59322,
            ]
        )
    elif Rv == 3.0:
        cor_vals = np.array(
            [
                5.55181,
                4.44232,
                3.62189,
                3.04890,
                2.71159,
                2.94688,
                3.30362,
                2.40863,
                1.93502,
            ]
        )
    elif Rv == 4.0:
        cor_vals = np.array(
            [
                3.07020,
                2.62711,
                2.29927,
                2.07040,
                1.94621,
                2.20610,
                2.45801,
                1.90003,
                1.60592,
            ]
        )
    elif Rv == 5.0:
        cor_vals = np.array(
            [
                1.58123,
                1.53798,
                1.50571,
                1.48330,
                1.48697,
                1.76164,
                1.95065,
                1.59486,
                1.40846,
            ]
        )
    elif Rv == 6.0:
        cor_vals = np.array(
            [
                0.588581,
                0.811898,
                0.976660,
                1.09190,
                1.18082,
                1.46533,
                1.61241,
                1.39142,
                1.27682,
            ]
        )
    else:
        cor_vals = np.array([0.0])

    return (x, cor_vals)


@pytest.mark.parametrize("Rv", [2.0, 3.0, 3.1, 4.0, 5.0, 6.0])
def test_extinction_GCC09_values(Rv):
    # get the correct values
    x, cor_vals = get_axav_cor_vals(Rv)

    # initialize extinction model
    tmodel = GCC09(Rv=Rv)

    # test
    np.testing.assert_allclose(tmodel(x), cor_vals, rtol=1e-5)


x_vals, axav_vals = get_axav_cor_vals(3.1)
test_vals = zip(x_vals, axav_vals)


@pytest.mark.parametrize("test_vals", test_vals)
def test_extinction_GCC09_single_values(test_vals):
    x, cor_val = test_vals

    # initialize extinction model
    tmodel = GCC09()

    # test
    np.testing.assert_allclose(tmodel(x), cor_val, rtol=1e-5)
    np.testing.assert_allclose(tmodel.evaluate(x, 3.1), cor_val, rtol=1e-5)



import pytest
import numpy as np
import astropy.units as u

from .helpers import grain_models


@pytest.mark.parametrize("model_class", grain_models)
def test_corvals(model_class):
    # instantiate extinction model
    tmodel = model_class()
    # get the possible list of models
    possmodels = tmodel.possnames.keys()

    for cmodel in possmodels:
        tmodel = model_class(cmodel)

        # test array evaluation
        x_vals = tmodel.data_x / u.micron
        y_vals = tmodel.data_axav
        tol = tmodel.data_tolerance
        np.testing.assert_allclose(tmodel(x_vals), y_vals, rtol=tol)


@pytest.mark.parametrize("model_class", grain_models)
def test_extinguish_values_Av_or_Ebv(model_class):
    # just test the default model
    ext = model_class()
    x = np.arange(ext.x_range[0], ext.x_range[1]) / u.micron
    cor_axav = ext(x)

    # calculate the cor_vals in fractional units
    Av = 1.0
    cor_vals_av = np.power(10.0, -0.4 * (cor_axav * Av))
    Ebv = 1.0
    Av_ebv = Ebv * ext.Rv
    cor_vals_ebv = np.power(10.0, -0.4 * cor_axav * Av_ebv)

    # tests
    np.testing.assert_equal(ext.extinguish(x, Av=Av), cor_vals_av)
    np.testing.assert_equal(ext.extinguish(x, Ebv=Ebv), cor_vals_ebv)



import numpy as np
import pytest

import astropy.units as u

from ..parameter_averages import VCG04


def get_axav_cor_vals(Rv):
    # testing wavenumbers
    x = np.array([8.0, 7.0, 6.0, 5.0, 4.6, 4.0, 3.4])

    # add units
    x = x / u.micron

    # correct values
    # from IDL version
    if Rv == 3.1:
        cor_vals = np.array(
            [3.36528, 2.84166, 2.58283, 2.88248, 3.25880, 2.43315, 2.00025]
        )
    elif Rv == 2.0:
        cor_vals = np.array(
            [5.20767, 4.25652, 3.74640, 4.16150, 4.73050, 3.33399, 2.54668]
        )
    elif Rv == 3.0:
        cor_vals = np.array(
            [3.47694, 2.92741, 2.65335, 2.96000, 3.34799, 2.48775, 2.03337]
        )
    elif Rv == 4.0:
        cor_vals = np.array(
            [2.61157, 2.26285, 2.10683, 2.35925, 2.65674, 2.06463, 1.77671]
        )
    elif Rv == 5.0:
        cor_vals = np.array(
            [2.09235, 1.86411, 1.77892, 1.99880, 2.24199, 1.81076, 1.622711]
        )
    elif Rv == 6.0:
        cor_vals = np.array(
            [1.74620, 1.59829, 1.56031, 1.75850, 1.96549, 1.64151, 1.52005]
        )
    else:
        cor_vals = np.array([0.0])

    return (x, cor_vals)


@pytest.mark.parametrize("Rv", [2.0, 3.0, 3.1, 4.0, 5.0, 6.0])
def test_extinction_VCG04_values(Rv):
    # get the correct values
    x, cor_vals = get_axav_cor_vals(Rv)

    # initialize extinction model
    tmodel = VCG04(Rv=Rv)

    # test
    np.testing.assert_allclose(tmodel(x), cor_vals, rtol=1e-5)


x_vals, axav_vals = get_axav_cor_vals(3.1)
test_vals = zip(x_vals, axav_vals)


@pytest.mark.parametrize("test_vals", test_vals)
def test_extinction_VCG04_single_values(test_vals):
    x, cor_val = test_vals

    # initialize extinction model
    tmodel = VCG04()

    # test
    np.testing.assert_allclose(tmodel(x), cor_val, rtol=1e-5)
    np.testing.assert_allclose(tmodel.evaluate(x, 3.1), cor_val, rtol=1e-5)
